from focs._tech import *

Category(
    name="LEARNING_CATEGORY",
    graphic="learning.png",
    colour=(54, 202, 229),
)

Category(
    name="GROWTH_CATEGORY",
    graphic="growth.png",
    colour=(116, 225, 107),
)

Category(
    name="PRODUCTION_CATEGORY",
    graphic="production.png",
    colour=(240, 106, 106, 255),
)

Category(
    name="CONSTRUCTION_CATEGORY",
    graphic="construction.png",
    colour=(241, 233, 87, 255),
)

Category(
    name="DEFENSE_CATEGORY",
    graphic="defense.png",
    colour=(70, 80, 215, 255),
)

Category(
    name="SHIP_HULLS_CATEGORY",
    graphic="ship_hulls.png",
    colour=(255, 139, 85, 255),
)

Category(
    name="SHIP_PARTS_CATEGORY",
    graphic="ship_parts.png",
    colour=(255, 139, 85, 255),
)

Category(
    name="SHIP_WEAPONS_CATEGORY",
    graphic="ship_weapons.png",
    colour=(255, 139, 85, 255),
)

Category(
    name="SPY_CATEGORY",
    graphic="spy.png",
    colour=(168, 0, 255, 255),
)



from focs._effects import (
    Abs,
    Adequate,
    Capital,
    Contains,
    CurrentTurn,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    GameRule,
    GasGiantType,
    Good,
    HasSpecies,
    HasTag,
    Homeworld,
    Hostile,
    Huge,
    InSystem,
    IsBuilding,
    IsSource,
    JumpsBetween,
    Large,
    LocalCandidate,
    MaxOf,
    Medium,
    Min,
    MinOf,
    NamedReal,
    OwnedBy,
    Planet,
    Poor,
    Population,
    ResourceSupplyConnected,
    SetHappiness,
    SetTargetHappiness,
    Small,
    Source,
    SpeciesDislikes,
    SpeciesLikes,
    Statistic,
    Target,
    Tiny,
    Uninhabitable,
    Unowned,
    Value,
)
from macros.misc import DESCRIPTION_EFFECTSGROUP_MACRO
from macros.opinion import POLICY_DISLIKE_SCALING
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

_STABILITY_PER_LIKED_FOCUS = 2.0

# Policy Liberty doubles de focus dislike effect, policy Conformance halves it, macro is in common/opinion.macros
_STABILITY_PER_DISLIKED_FOCUS = 2.0 * POLICY_DISLIKE_SCALING

_ENVIRONMENT_STABILITY_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Uninhabitable])
        & (GameRule(type=int, name="RULE_HOSTILE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="UNINHABTIABLE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HOSTILE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Hostile])
        & (GameRule(type=int, name="RULE_HOSTILE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="HOSTILE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HOSTILE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Poor])
        & (GameRule(type=int, name="RULE_POOR_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="POOR_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_POOR_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Adequate])
        & (GameRule(type=int, name="RULE_ADEQUATE_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="ADEQUATE_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_ADEQUATE_ENVIRONMENT_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(environment=[Good])
        & (GameRule(type=int, name="RULE_GOOD_ENVIRONMENT_STABILITY") != 0),
        accountinglabel="GOOD_ENVIRONMENT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_GOOD_ENVIRONMENT_STABILITY")),
    ),
]

_PLANET_SIZE_STABILITY_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Tiny]) & (GameRule(type=int, name="RULE_TINY_SIZE_STABILITY") != 0),
        accountinglabel="TINY_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_TINY_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Small]) & (GameRule(type=int, name="RULE_SMALL_SIZE_STABILITY") != 0),
        accountinglabel="SMALL_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_SMALL_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Medium]) & (GameRule(type=int, name="RULE_MEDIUM_SIZE_STABILITY") != 0),
        accountinglabel="SMALL_MEDIUM_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_MEDIUM_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Large]) & (GameRule(type=int, name="RULE_LARGE_SIZE_STABILITY") != 0),
        accountinglabel="LARGE_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_LARGE_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(size=[Huge]) & (GameRule(type=int, name="RULE_HUGE_SIZE_STABILITY") != 0),
        accountinglabel="HUGE_PLANET_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_HUGE_SIZE_STABILITY")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet()
        & Planet(type=[GasGiantType])
        & (GameRule(type=int, name="RULE_GAS_GIANT_SIZE_STABILITY") != 0),
        accountinglabel="GAS_GIANT_STABILITY_LABEL",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_GAS_GIANT_SIZE_STABILITY")),
    ),
]

_COMMON_HAPPINESS_EFFECTS = [
    *_ENVIRONMENT_STABILITY_MODIFIER,
    *_PLANET_SIZE_STABILITY_MODIFIER,
    EffectsGroup(  # increase or decrease 1 per turn towards target
        scope=IsSource,
        activation=Planet() & (LocalCandidate.LastTurnConquered < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetHappiness(
            value=Value
            + MinOf(float, Abs(float, Value(Target.TargetHappiness) - Value), 1)
            * (1 - 2 * (Value > Value(Target.TargetHappiness)))
        ),
    ),
    EffectsGroup(  # artistic species make other planets with different artistic species on them and in the same system system more stable
        scope=Planet()
        & InSystem(id=Source.SystemID)
        & ~IsSource
        & HasSpecies()
        & ~HasSpecies(name=[Source.Species])
        & HasTag(name="ARTISTIC")
        # (SpeciesSpeciesOpinion species = LocalCandidate.Species species = Source.Species > 0) # TODO: implement within an Or condition, once species-species opinions are working...
        ,
        activation=Planet() & HasSpecies() & HasTag(name="ARTISTIC"),
        accountinglabel="ARTISAN_APPRECIATION",
        effects=SetTargetHappiness(value=Value + 1.0),
    ),
    EffectsGroup(  # when not owned by an empire, directly apply baseline stability
        scope=IsSource,
        activation=GameRule(type=int, name="RULE_BASELINE_PLANET_STABILITY") != 0,
        accountinglabel="RULE_BASELINE_PLANET_STABILITY",
        effects=SetTargetHappiness(value=Value + GameRule(type=float, name="RULE_BASELINE_PLANET_STABILITY")),
    ),
    EffectsGroup(  # more stable when liked focuses are adopted
        scope=IsSource,
        activation=Population(low=0.001) & SpeciesLikes(name=LocalCandidate.Focus),
        accountinglabel="LIKES_FOCUS_LABEL",
        effects=SetTargetHappiness(value=Value + _STABILITY_PER_LIKED_FOCUS),
    ),
    EffectsGroup(  # less stable when disliked focuses are adopted
        scope=IsSource,
        activation=~Unowned & Population(low=0.001) & SpeciesDislikes(name=LocalCandidate.Focus),
        accountinglabel="DISLIKES_FOCUS_LABEL",
        effects=SetTargetHappiness(value=Value - _STABILITY_PER_DISLIKED_FOCUS),
    ),
    EffectsGroup(  # species on their own homeworld are more stable
        scope=IsSource,
        activation=Homeworld(name=[Source.Species]),
        accountinglabel="HOMEWORLD_BONUS",
        effects=SetTargetHappiness(value=Value + 5),
    ),
    EffectsGroup(  # capital planet is more stable
        scope=IsSource,
        activation=Capital,
        accountinglabel="CAPITAL_LABEL",
        effects=SetTargetHappiness(value=Value + 10),
    ),
    EffectsGroup(  # concentration camp nullifies stability
        scope=IsSource,
        activation=Contains(IsBuilding(name=["BLD_CONC_CAMP"])),
        accountinglabel="CONCENTRATION_CAMPS_LABEL",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=[SetHappiness(value=0), SetTargetHappiness(value=0)],
    ),
]

# this should ensure that there is a capital / admin for JUMPS_TO_CAPITAL_OR_REGAD
_SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD = ResourceSupplyConnected(
    empire=Source.Owner, condition=(Capital | IsBuilding(name=["BLD_REGIONAL_ADMIN"])) & OwnedBy(empire=Source.Owner)
)


def _jumps_to_capital_or_regad(type):
    return Statistic(
        type,
        Min,
        # Replace JumpsBetween by JumpsBetweenByEmpireSupplyConnections once it is implemented
        value=JumpsBetween(Source.ID, LocalCandidate.ID),
        condition=(
            (Capital | IsBuilding(name=["BLD_REGIONAL_ADMIN"]))
            & OwnedBy(empire=Source.Owner)
            & ResourceSupplyConnected(empire=Source.Owner, condition=IsSource)
        ),
    )


_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY = [
    EffectsGroup(  # close supply connected to capital planet is more stable
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & _SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD
        & (_jumps_to_capital_or_regad(int) < 5),
        accountinglabel="CAPITAL_CONNECTION_LABEL",
        effects=SetTargetHappiness(value=Value + MinOf(float, 5, 5 - _jumps_to_capital_or_regad(float))),
    ),
    EffectsGroup(  # far supply connected to capital planet is less
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_CONFEDERATION")
        & _SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD
        & (_jumps_to_capital_or_regad(int) > 5),
        accountinglabel="CAPITAL_POOR_CONNECTION_LABEL",
        effects=SetTargetHappiness(value=Value + MaxOf(float, -5, 5 - _jumps_to_capital_or_regad(float))),
    ),
    EffectsGroup(  # no connection is much less stable
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~Capital
        & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_CONFEDERATION")
        & ~_SUPPLY_CONNECTED_TO_CAPITAL_OR_REGAD,
        accountinglabel="CAPITAL_DISCONNECTION_LABEL",
        effects=SetTargetHappiness(
            value=Value - NamedReal(name="DISCONNECTED_FROM_CAPITAL_AND_REGIONAL_ADMIN_STABILITY_PENALTY", value=10)
        ),
    ),
]

VERY_BAD_HAPPINESS = [
    *_COMMON_HAPPINESS_EFFECTS,
    *_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
    EffectsGroup(
        description="VERY_BAD_HAPPINESS_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="VERY_BAD_HAPPINESS_LABEL",
        effects=SetTargetHappiness(value=Value + NamedReal(name="VERY_BAD_HAPPINESS_VAL", value=-5.0)),
    ),
]

BAD_HAPPINESS = [
    *_COMMON_HAPPINESS_EFFECTS,
    *_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
    EffectsGroup(
        description="BAD_HAPPINESS_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="BAD_HAPPINESS_LABEL",
        effects=SetTargetHappiness(value=Value + NamedReal(name="BAD_HAPPINESS_VAL", value=-2.5)),
    ),
]

AVERAGE_HAPPINESS = [
    *_COMMON_HAPPINESS_EFFECTS,
    *_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
]

GOOD_HAPPINESS = [
    *_COMMON_HAPPINESS_EFFECTS,
    *_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
    EffectsGroup(
        description="GOOD_HAPPINESS_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GOOD_HAPPINESS_LABEL",
        effects=SetTargetHappiness(value=Value + NamedReal(name="GOOD_HAPPINESS_VAL", value=2.5)),
    ),
]

GREAT_HAPPINESS = [
    *_COMMON_HAPPINESS_EFFECTS,
    *_STANDARD_SPECIES_CAPITAL_SUPPLY_CONNECTION_STABILITY,
    EffectsGroup(
        description="GREAT_HAPPINESS_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GREAT_HAPPINESS_LABEL",
        effects=SetTargetHappiness(value=Value + NamedReal(name="GREAT_HAPPINESS_VAL", value=5)),
    ),
]


INDEPENDENT_HAPPINESS = [DESCRIPTION_EFFECTSGROUP_MACRO("INDEPENDENT_HAPPINESS_DESC"), *_COMMON_HAPPINESS_EFFECTS]



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_CAMO_AST_HULL",
    description="SHP_CAMO_AST_HULL_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=52 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["SHP_ASTEROID_HULLS"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_SMALL_CAMOUFLAGE_ASTEROID"),
        Item(type=UnlockShipHull, name="SH_CAMOUFLAGE_ASTEROID"),
    ],
    graphic="icons/ship_hulls/camouflage_asteroid_hull_small.png",
)



"""
This module contains tools to work with pre-aggregated information about AI state.
This information is not stored in the save game, so it recalculated each turn.

The main purpose of this module is to separate code that aggregates information and decision-making code.
Decision-making code is smart, and this is the dumb module,
which means that we will try to make this module API to cover all needs of decision-making code, in order to reduce extra manipulation.
If getting information requires more than one line it might be a candidate to a function in this module.

Module API:
Module API is represented as several functions.
All public functions are just lazy objects.
They are initialized upon the first request each turn, and the value is then cached until the end of that turn.

Avoiding classes that are exposed to outer scope requires some additional calls inside functions,
 but this price is not too high to get a simple API.

The module is split into sub-modules, based on the idea that each submodule should work with the same set of data.
This allows achieving the following: each submodule is relatively small and is therefore easy to understand and extend.

What should be put to the module:
If your code works with freeOrionAIInterface only and data can be queried each time, this code is a candidate for this module.

How to organize data inside that module:
Simple rules: keep files small and dependencies between sub-modules under control.
If you don't sure which sub-modules should be chosen - consider creating a new one.
If your code adds new sub-modules dependency to existing sub-modules - consider creating a new one.
"""

from turn_state._empire_resources import (
    computronium_candidates,
    have_asteroids,
    have_computronium,
    have_gas_giant,
    have_honeycomb,
    have_nest,
    have_ruins,
    have_worldtree,
    honeycomb_candidates,
    luxury_resources,
    owned_asteroid_coatings,
    population_with_industry_focus,
    set_have_asteroids,
    set_have_gas_giant,
    set_have_nest,
)
from turn_state._planet_state import (
    get_all_empire_planets,
    get_colonized_planets,
    get_colonized_planets_in_system,
    get_empire_outposts,
    get_empire_planets_by_species,
    get_empire_planets_with_species,
    get_inhabited_planets,
    get_number_of_colonies,
    get_owned_planets,
    get_owned_planets_in_system,
    get_unowned_empty_planets,
)
from turn_state._supply_state import (
    get_distance_to_enemy_supply,
    get_supply_group,
    get_supply_group_id,
    get_system_supply,
    get_systems_by_supply_tier,
    supply_connected,
)



import freeOrionAIInterface as fo
from logging import warning
from typing import Optional

from AIDependencies import CombatTarget
from aistate_interface import get_aistate
from CombatRatingsAI._targets import get_allowed_targets
from common.fo_typing import AttackCount, AttackDamage, ShipId
from freeorion_tools import get_ship_part
from freeorion_tools.caching import cache_for_current_turn


class ShipCombatStats:
    """Stores all relevant stats of a ship for combat strength evaluation."""

    def __init__(
        self,
        *,
        attacks: Optional[dict[AttackDamage, AttackCount]] = None,
        structure=1.0,
        shields=0.0,
        fighter_capacity=0,
        fighter_launch_rate=0,
        fighter_damage=0,
        flak_shots=0,
        has_interceptors=False,
        damage_vs_planets=0,
        has_bomber=False,
    ):
        self._structure = structure
        self.shields = shields
        self.attacks: dict[AttackDamage, AttackCount] = {} if attacks is None else attacks

        self._fighter_capacity = fighter_capacity
        self._fighter_launch_rate = fighter_launch_rate
        self._fighter_damage = fighter_damage

        self._flak_shots = flak_shots
        self._has_interceptors = has_interceptors

        self._damage_vs_planets = damage_vs_planets
        self._has_bomber = has_bomber

    def __getstate__(self):
        return {
            "_structure": self._structure,
            "shields": self.shields,
            "attacks": self.attacks,
            "_fighter_capacity": self._fighter_capacity,
            "_fighter_launch_rate": self._fighter_launch_rate,
            "_fighter_damage": self._fighter_damage,
            "_flak_shots": self._flak_shots,
            "_has_interceptors": self._has_interceptors,
            "_damage_vs_planets": self._damage_vs_planets,
            "_has_bomber": self._has_bomber,
        }

    def __hash__(self):
        return hash((tuple(self.attacks.items()), self._structure, self.shields))

    def __eq__(self, other):
        return self.__getstate__() == other.__getstate__()

    def __str__(self):
        return str(self.__getstate__())

    def get_rating(self, enemy_stats: "ShipCombatStats" = None) -> float:
        """Calculate a rating against specified enemy.

        If no enemy is specified, will rate against the empire standard enemy

        :param enemy_stats: Enemy stats to be rated against - if None
        :return: rating against specified enemy
        """

        # adjust base stats according to enemy stats
        # The fighter rating calculations are heavily based upon the enemy stats.
        # So, for now, we compare at least against a certain standard enemy.
        enemy_stats = enemy_stats or get_aistate().get_standard_enemy()

        my_hit_points = self._structure
        if enemy_stats:
            my_hit_points *= self._calculate_shield_factor(enemy_stats.attacks, self.shields)
            my_total_attack = sum(n * max(dmg - enemy_stats.shields, 0.001) for dmg, n in self.attacks.items())
        else:
            my_total_attack = sum(n * dmg for dmg, n in self.attacks.items())
            my_hit_points += self.shields

        my_total_attack += self._estimate_fighter_damage()
        # TODO: Consider enemy fighters
        return my_total_attack * my_hit_points

    def _calculate_shield_factor(self, e_attacks: dict[AttackDamage, AttackCount], my_shields: float) -> float:
        """
        Calculates shield factor based on enemy attacks and our shields.
        It is possible to have e_attacks with number attacks == 0,
        in that case we consider that there is an issue with the enemy stats and we jut set value to 1.0.
        """
        if not e_attacks:
            return 1.0
        e_total_attack = sum(n * dmg for dmg, n in e_attacks.items())
        if e_total_attack:
            e_net_attack = sum(n * max(dmg - my_shields, 0.001) for dmg, n in e_attacks.items())
            e_net_attack = max(e_net_attack, 0.1 * e_total_attack)
            shield_factor = e_total_attack / e_net_attack
            return max(1.0, shield_factor)
        else:
            return 1.0

    def _estimate_fighter_damage(self):
        if self._fighter_launch_rate == 0:
            return 0
        full_launch_bouts = self._fighter_capacity // self._fighter_launch_rate
        survival_rate = 0.2  # TODO estimate chance of a fighter not to be shot down in a bout
        flying_fighters = 0
        total_fighter_damage = 0
        # Cut that values down to a single turn (four bouts means max three launch bouts)
        num_bouts = fo.getGameRules().getInt("RULE_NUM_COMBAT_ROUNDS")
        for firing_bout in range(num_bouts - 1):
            if firing_bout < full_launch_bouts:
                flying_fighters = (flying_fighters * survival_rate) + self._fighter_launch_rate
            elif firing_bout == full_launch_bouts:
                # now handle a bout with lower capacity launch
                flying_fighters = (flying_fighters * survival_rate) + (
                    self._fighter_capacity % self._fighter_launch_rate
                )
            else:
                flying_fighters = flying_fighters * survival_rate
            total_fighter_damage += self._fighter_damage * flying_fighters
        return total_fighter_damage / num_bouts

    def get_rating_vs_planets(self) -> float:
        """Heuristic to estimate combat strength against planets"""
        damage = self._damage_vs_planets
        if self._has_bomber:
            damage += self._estimate_fighter_damage()
        return damage * (self._structure + self.shields)


@cache_for_current_turn
def get_ship_combat_stats(ship_id: ShipId, max_stats=False) -> ShipCombatStats:  # noqa: C901
    """
    Return combat stats for the ship.

    If max_stats is True use maximum values.
    """
    universe = fo.getUniverse()
    ship = universe.getShip(ship_id)
    if not ship:
        return ShipCombatStats()  # TODO: Add some estimate for stealthed ships

    if max_stats:
        structure = ship.initialMeterValue(fo.meterType.maxStructure)
        shields = ship.initialMeterValue(fo.meterType.maxShield)
    else:
        structure = ship.initialMeterValue(fo.meterType.structure)
        shields = ship.initialMeterValue(fo.meterType.shield)
    attacks: dict[AttackDamage, AttackCount] = {}
    fighter_launch_rate = 0
    fighter_capacity = 0
    fighter_damage = 0
    flak_shots = 0
    has_bomber = False
    has_interceptors = False
    damage_vs_planets = 0
    design = ship.design
    if design and (ship.isArmed or ship.hasFighters):
        meter_choice = fo.meterType.maxCapacity if max_stats else fo.meterType.capacity
        for partname in design.parts:
            if not partname:
                continue
            pc = get_ship_part(partname).partClass
            if pc == fo.shipPartClass.shortRange:
                allowed_targets = get_allowed_targets(partname)
                damage = AttackDamage(ship.currentPartMeterValue(meter_choice, partname))
                shots = int(ship.currentPartMeterValue(fo.meterType.secondaryStat, partname))
                if allowed_targets & CombatTarget.SHIP:
                    attacks[damage] = AttackCount(attacks.get(damage, 0) + shots)
                if allowed_targets & CombatTarget.FIGHTER:
                    flak_shots += 1
                if allowed_targets & CombatTarget.PLANET:
                    damage_vs_planets += damage * shots
            elif pc == fo.shipPartClass.fighterBay:
                fighter_launch_rate += ship.currentPartMeterValue(fo.meterType.capacity, partname)
            elif pc == fo.shipPartClass.fighterHangar:
                allowed_targets = get_allowed_targets(partname)
                # for hangars, capacity meter is already counting contributions from ALL hangars.
                fighter_capacity = ship.currentPartMeterValue(meter_choice, partname)
                part_damage = ship.currentPartMeterValue(fo.meterType.secondaryStat, partname)
                if part_damage != fighter_damage and fighter_damage > 0:
                    # the C++ code fails also in this regard, so FOCS content *should* not allow this.
                    # TODO: Depending on future implementation, might actually need to handle this case.
                    warning("Multiple hangar types present on one ship, estimates expected to be wrong.")
                if allowed_targets & CombatTarget.SHIP:
                    fighter_damage = max(fighter_damage, part_damage)
                if allowed_targets & CombatTarget.PLANET:
                    has_bomber = True
                if allowed_targets & CombatTarget.FIGHTER:
                    has_interceptors = True

    return ShipCombatStats(
        attacks=attacks,
        structure=structure,
        shields=shields,
        fighter_capacity=fighter_capacity,
        fighter_launch_rate=fighter_launch_rate,
        fighter_damage=fighter_damage,
        flak_shots=flak_shots,
        has_interceptors=has_interceptors,
        damage_vs_planets=damage_vs_planets,
        has_bomber=has_bomber,
    )



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_BIOTERROR",
    description="GRO_BIOTERROR_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_NANOTECH_MED"],
    unlock=Item(type=UnlockBuilding, name="BLD_BIOTERROR_PROJECTOR"),
    graphic="icons/tech/bioterror_facilities.png",
)



import itertools
import math
import random
import sys
from collections.abc import Iterable, Iterator
from hashlib import md5

error_list = []


def int_hash(s):
    h = md5()
    h.update(s)
    return int(h.hexdigest(), 16)


def seed_rng(seed):
    """
    Seeds the default RNG with the specified seed
    """
    random.seed(seed)
    # The following jumpahead call can be uncommented to work around this issue of the Mersenne Twister PRNG
    # (quoted from wikipedia http://en.wikipedia.org/wiki/Mersenne_twister):
    # "It can take a long time to start generating output that passes randomness tests, if the initial state is highly
    # non-random-- particularly if the initial state has many zeros. A consequence of this is that two instances of the
    # generator, started with initial states that are almost the same, will usually output nearly the same sequence
    # for many iterations, before eventually diverging."
    # random.jumpahead(999999)


def distance(start, end):
    """
    Calculates linear distance between two coordinates.
    """
    x1, y1 = start
    x2, y2 = end
    return math.hypot(float(x1) - float(x2), float(y1) - float(y2))


def report_error(msg):
    """
    Handles error messages.
    """
    error_list.append(msg)
    print(msg, file=sys.stderr)


def unique_product(first: Iterable[int], second: Iterable[int]) -> Iterator[tuple[int, int]]:
    """
    Create all possible unique pairs for two iterables.
    Both iterables should consist of comparable objects.
    Pair is sorted in ascending order.
    """
    return filter(lambda x: x[0] < x[1], itertools.product(first, second))


class MapGenerationError(RuntimeError):
    """Map generation runtime error."""



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_MONOMOLEC_LATTICE",
    description="SHP_MONOMOLEC_LATTICE_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=1000 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["SHP_ASTEROID_REFORM"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_CRYSTALLIZED_ASTEROID"),
        Item(type=UnlockShipPart, name="AR_CRYSTAL_PLATE"),
    ],
    graphic="icons/ship_parts/crystal_plating.png",
)



from focs._species import *
from species.species_macros.detection import GOOD_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TUNDRA_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import INDEPENDENT_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GREAT_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS

Species(
    name="SP_FURTHEST",
    description="SP_FURTHEST_DESC",
    gameplay_description="SP_FURTHEST_GAMEPLAY_DESC",
    native=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "BAD_RESEARCH",
        "BAD_POPULATION",
        "BAD_INFLUENCE",
        "BAD_SUPPLY",
        "GOOD_DETECTION",
        "GREAT_STEALTH",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "INDEPENDENT_HAPPINESS",
        "PEDIA_INDEPENDENT",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_GROWTH",
    likes=[
        "FOCUS_GROWTH",
        "SHIMMER_SILK_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "FRUIT_SPECIAL",
        "FORTRESS_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_CHECKPOINTS",
        "PLC_ENVIRONMENTALISM",
        "PLC_ISOLATION",
        "PLC_NO_SUPPLY",
    ],
    dislikes=[
        "BLD_STOCKPILING_CENTER",
        "BLD_MEGALITH",
        "BLD_REGIONAL_ADMIN",
        "BLD_SPACE_ELEVATOR",
        "BLD_STARGATE",
        "BLD_PLANET_CLOAK",
        "WORLDTREE_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "MONOPOLE_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_CAPITAL_MARKETS",
        "PLC_COLONIALISM",
        "PLC_INTERSTELLAR_INFRA",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *BAD_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *INDEPENDENT_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_FURTHEST"),
        *BAD_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        *GOOD_DETECTION,
        *GREAT_STEALTH,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TUNDRA_STANDARD_EP,
    graphic="icons/species/furthest.png",
)



from focs._effects import (
    Destroy,
    EffectsGroup,
    GenerateSitRepMessage,
    IsBuilding,
    IsSource,
    LocalCandidate,
    MaxOf,
    Object,
    OwnedBy,
    Partial,
    Planet,
    Population,
    SetOwner,
    SetVisibility,
    Source,
    Target,
    ThisBuilding,
    Turn,
    ValueVisibility,
)
from macros.misc import UNOWNED_EMPIRE_ID
from macros.priorities import POPULATION_OVERRIDE_PRIORITY

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_ABANDON_OUTPOST",
    description="BLD_ABANDON_OUTPOST_DESC",
    captureresult=DestroyOnCapture,  # type: ignore[reportUnboundVariable]
    buildcost=20,
    buildtime=1,
    location=Planet() & OwnedBy(empire=Source.Owner) & Population(high=0),
    effectsgroups=[
        # PREPARE FOR TWO TURNS
        EffectsGroup(
            scope=IsSource,
            activation=Turn(high=Source.CreationTurn + 2),
            stackinggroup="ABANDON_OUTPOST_STACK",
            effects=[
                GenerateSitRepMessage(
                    message="SITREP_OUTPOST_ABANDONED_PREPARATION",
                    label="SITREP_OUTPOST_ABANDONED_PREPARATION_LABEL",
                    icon="icons/tech/environmental_encapsulation.png",
                    parameters={"planet": Source.PlanetID, "rawtext": Source.CreationTurn + 4},
                    empire=Source.Owner,
                )
            ],
        ),
        # THEN CANT ABANDON IF BATTLE
        EffectsGroup(
            scope=IsSource,
            activation=Turn(low=Source.CreationTurn + 3, high=Source.System.LastTurnBattleHere),
            stackinggroup="ABANDON_OUTPOST_STACK",
            effects=[
                GenerateSitRepMessage(
                    message="SITREP_OUTPOST_ABANDONED_PREVENTED_BY_BATTLE",
                    label="SITREP_OUTPOST_ABANDONED_LABEL",
                    icon="icons/tech/environmental_encapsulation.png",
                    parameters={"planet": Target.PlanetID, "system": Target.SystemID},
                    empire=Source.Owner,
                )
            ],
        ),
        # ELSE ABANDON OUTPOST
        EffectsGroup(
            scope=Object(id=Source.PlanetID) & Planet(),
            activation=Turn(low=MaxOf(int, Source.System.LastTurnBattleHere + 1, Source.CreationTurn + 3)),
            stackinggroup="ABANDON_OUTPOST_STACK",
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=[
                SetOwner(empire=UNOWNED_EMPIRE_ID),
                GenerateSitRepMessage(
                    message="SITREP_OUTPOST_ABANDONED",
                    label="SITREP_OUTPOST_ABANDONED_LABEL",
                    icon="icons/tech/environmental_encapsulation.png",
                    parameters={"planet": Target.ID},
                    empire=Source.Owner,
                ),
                SetVisibility(empire=Source.Owner, visibility=MaxOf("Visibility", ValueVisibility, Partial)),
            ],
        ),
        EffectsGroup(
            scope=(LocalCandidate.PlanetID == Source.PlanetID) & IsBuilding() & (~IsBuilding(name=[ThisBuilding])),
            activation=Turn(low=MaxOf(int, Source.System.LastTurnBattleHere + 1, Source.CreationTurn + 3)),
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=[SetOwner(empire=UNOWNED_EMPIRE_ID)],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=Turn(low=MaxOf(int, Source.System.LastTurnBattleHere + 1, Source.CreationTurn + 3)),
            effects=Destroy,
        ),
    ],
    icon="icons/tech/environmental_encapsulation.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_INDUSTRY_CENTER_II",
    description="PRO_INDUSTRY_CENTER_II_DESC",
    short_description="BUILDING_REFINE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_INDUSTRY_CENTER_I"],
    graphic="icons/tech/industrial_centre_ii.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SPACE_FLUX_COMPOSITION",
    description="SHP_SPACE_FLUX_COMPOSITION_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_FLUX_HULL_TECHS"],
    prerequisites=["SHP_SPACE_FLUX_BUBBLE", "SHP_SPACE_FLUX_DRIVE"],
    unlock=Item(type=UnlockShipHull, name="SH_SPACE_FLUX_COMPOSITE"),
    # FIXME graphic
    graphic="icons/ship_hulls/agregate_asteroid_hull_small.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SPY_LIGHTHOUSE",
    description="SPY_LIGHTHOUSE_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_SPY_CATEGORY"],
    prerequisites=["SPY_DETECT_2"],
    unlock=Item(type=UnlockBuilding, name="BLD_LIGHTHOUSE"),
    graphic="icons/tech/interstellar_lighthouse.png",
)



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    Focus,
    IsSource,
    LocalCandidate,
    MinOf,
    Planet,
    SetMaxShield,
    SetShield,
    Target,
    Unowned,
    Value,
)
from macros.misc import PLANET_SHIELD_FACTOR
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
)


def _native_planetary_shields(strength: float):
    return EffectsGroup(
        scope=IsSource,
        activation=Planet() & Unowned,
        accountinglabel="NATIVE_PLANETARY_SHIELDS_LABEL",
        effects=(SetMaxShield(value=Value + strength)),
    )


AVERAGE_PLANETARY_SHIELDS = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
        stackinggroup="FOCUS_PROTECTION_SHIELDS_STACK",
        accountinglabel="FOCUS_PROTECTION_LABEL",
        priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
        effects=SetMaxShield(value=Value * 2),
    ),
    EffectsGroup(  # increase 1 per turn, up to max
        scope=IsSource,
        activation=Planet()
        & Unowned
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        accountinglabel="DEF_ROOT_DEFENSE",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetShield(value=MinOf(float, Value(Target.MaxShield), Value + PLANET_SHIELD_FACTOR)),
    ),
]



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ART_HEAVENLY",
    description="CON_ART_HEAVENLY_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=150 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_CONTGRAV_ARCH", "SHP_CONTGRAV_MAINT"],
    unlock=Item(type=UnlockBuilding, name="BLD_ART_MOON"),
    graphic="icons/tech/artificial_heavenly_bodies.png",
)



from abc import ABC
from typing import Callable, TypeVar

Serializable = TypeVar("Serializable")


class Serializer(ABC):
    def __init__(
        self,
        *,
        serializer: Callable[[Serializable], str],
        deserializer: Callable[[str], Serializable],
    ):
        self._deserializer = deserializer
        self._serializer = serializer

    def deserialize(self, value: str) -> Serializable:
        try:
            return self._deserializer(value)
        except Exception as e:
            raise ValueError(f"{self.__class__.__name__} fail to deserialize {value}: {e}") from e

    def serialize(self, value: Serializable) -> str:
        return self._serializer(value)


class _IntSerializer(Serializer):
    def __init__(self):
        super().__init__(serializer=str, deserializer=int)


class _FloatSerializer(Serializer):
    def __init__(self):
        super().__init__(serializer=str, deserializer=float)


class _StrSerializer(Serializer):
    def __init__(self):
        super().__init__(serializer=str, deserializer=str.strip)


class ListSerializer(Serializer):
    def __init__(self, item_serializer: Serializer):
        def serializer(value: list[Serializable]) -> str:
            return ", ".join(item_serializer.serialize(val) for val in value)

        def deserializer(serialized: str) -> Serializable:
            elements = serialized.split(",")
            return [item_serializer.deserialize(v) for v in elements]

        super().__init__(serializer=serializer, deserializer=deserializer)


class DictSerializer(Serializer):
    def __init__(self, items: dict[str, Serializer]):
        def serializer(value: dict[str, Serializable]) -> str:
            assert set(value) == set(items), f"Keys should match {sorted(value)} != {sorted(items)}"
            serialized_dict = {}
            for key, val in value.items():  # preserve order of values dict
                serializer_instance = items[key]
                serialized_dict[key] = serializer_instance.serialize(val)

            return ", ".join(f"{k}: {v}" for k, v in serialized_dict.items())

        def deserializer(serialized: str) -> Serializable:
            elements = serialized.split(",")
            assert len(elements) == len(items)

            def extract_pair(pair_string: str):
                key, val = pair_string.split(":")
                return key.strip(), val.strip()

            pairs = (extract_pair(x) for x in elements)
            return {k: items[k].deserialize(v) for (k, v) in pairs}

        super().__init__(serializer=serializer, deserializer=deserializer)


to_int = _IntSerializer()
to_str = _StrSerializer()
to_float = _FloatSerializer()



import pytest

from freeorion_tools.bonus_calculation import Bonus, adjust_direction


@pytest.mark.parametrize(
    ("required_stability", "actual_stability", "expected"),
    ((0.0, 1.0, 5.5), (1.0, 1.0, 5.5), (2.0, 1.0, 0)),
)
def test_get_bonus_with_available_depends_on_stability(required_stability, actual_stability, expected):
    bonus = Bonus(True, required_stability, 5.5)
    assert bonus.get_bonus(actual_stability) == expected


@pytest.mark.parametrize(
    ("required_stability", "actual_stability"),
    ((0.0, 1.0), (1.0, 1.0), (2.0, 1.0)),
)
def test_get_bonus_with_not_available_is_zero(required_stability, actual_stability):
    bonus = Bonus(False, required_stability, 5.5)
    assert bonus.get_bonus(actual_stability) == 0.0


@pytest.mark.parametrize(
    ("threshold", "current", "target", "expected"),
    ((10.0, 9.0, 6.0, 0), (10.0, 8.0, 9.0, 0), (9.0, 6.5, 9.0, 1), (10.0, 12.0, 9.9, -1)),
)
def test_adjust_direction(threshold, current, target, expected):
    assert adjust_direction(threshold, current, target) == expected



from focs._effects import (
    BlackHole,
    EffectsGroup,
    Floor,
    InSystem,
    MaxOf,
    NamedReal,
    Neutron,
    NoStar,
    OwnedBy,
    Red,
    SetStealth,
    Ship,
    Source,
    Star,
    StatisticCount,
    Target,
    Value,
)
from focs._tech import *

Tech(
    name="SPY_ROOT_DECEPTION",
    description="SPY_ROOT_DECEPTION_DESC",
    short_description="THEORY_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_SPY_CATEGORY", "THEORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner) & Star(type=[NoStar]),
            accountinglabel="SPY_DECEPTION_EMPTY_SPACE_PENALTY",
            effects=SetStealth(value=Value + NamedReal(name="SPY_DECEPTION_EMPTY_SPACE_PENALTY", value=-10.0)),
        ),
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner) & Star(type=[Red]),
            accountinglabel="SPY_DECEPTION_DIM_STAR_PENALTY",
            effects=SetStealth(value=Value + NamedReal(name="SPY_DECEPTION_DIM_STAR_PENALTY", value=-5.0)),
        ),
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner) & Star(type=[Neutron]),
            accountinglabel="SPY_DECEPTION_SUBSTELLAR_INTERFERENCE",
            effects=SetStealth(value=Value + NamedReal(name="SPY_DECEPTION_NEUTRON_INTERFERENCE", value=5.0)),
        ),
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner) & Star(type=[BlackHole]),
            accountinglabel="SPY_DECEPTION_SUBSTELLAR_INTERFERENCE",
            effects=SetStealth(value=Value + NamedReal(name="SPY_DECEPTION_BLACK_INTERFERENCE", value=10.0)),
        ),
        EffectsGroup(
            scope=Ship & InSystem() & OwnedBy(empire=Source.Owner),
            accountinglabel="FLEET_UNSTEALTHINESS",
            effects=SetStealth(
                value=Value
                - NamedReal(name="FLEET_UNSTEALTH_SHIPS_SCALING", value=5.0)
                * Floor(
                    float,
                    MaxOf(
                        float,
                        0,
                        (
                            StatisticCount(
                                float, condition=Ship & InSystem(id=Target.SystemID) & OwnedBy(empire=Source.Owner)
                            )
                        )
                        - NamedReal(name="FLEET_UNSTEALTHY_SHIPS_THRESHOLD", value=10),
                    )
                    ** 0.5,
                )
            )
            # large fleets only start affecting stealth when there are more than the threshold of ships in a single system
        ),
    ],
)



import freeOrionAIInterface as fo
from logging import info

import AIDependencies
import AIstate
from AIDependencies import Tags
from aistate_interface import get_aistate
from colonization import rate_piloting, special_is_nest
from colonization.calculate_planet_colonization_rating import empire_metabolisms
from colonization.calculate_population import active_growth_specials
from common.listeners import register_pre_handler
from common.print_utils import Bool, Sequence, Table, Text
from empire.buildings_locations import set_building_locations
from empire.colony_builders import can_build_colony_for_species, set_colony_builders
from empire.colony_status import (
    set_colonies_is_under_attack,
    set_colonies_is_under_treat,
)
from empire.growth_specials import set_growth_special
from empire.pilot_rating import set_pilot_rating_for_planet
from empire.ship_builders import get_ship_builder_locations, set_ship_builders
from empire.survey_lock import survey_universe_lock
from EnumsAI import FocusType
from freeorion_tools import get_partial_visibility_turn, get_species_tag_grade
from freeorion_tools.timers import AITimer
from turn_state import (
    get_empire_planets_by_species,
    get_owned_planets,
    set_have_asteroids,
    set_have_gas_giant,
    set_have_nest,
)

register_pre_handler("generateOrders", survey_universe_lock.lock)

survey_timer = AITimer("empire.surver_universe()")


def survey_universe():  # noqa: C901
    survey_timer.start("Categorizing Visible Planets")
    universe = fo.getUniverse()
    empire_id = fo.empireID()
    current_turn = fo.currentTurn()

    AIstate.empireStars.clear()
    empire_metabolisms.clear()
    active_growth_specials.clear()

    # var setup done
    aistate = get_aistate()
    for sys_id in universe.systemIDs:
        system = universe.getSystem(sys_id)
        if not system:
            continue
        local_ast = False
        local_gg = False
        empire_has_qualifying_planet = False
        if sys_id in AIstate.colonyTargetedSystemIDs:
            empire_has_qualifying_planet = True
        for pid in system.planetIDs:
            planet = universe.getPlanet(pid)
            if not planet:
                continue
            if planet.size == fo.planetSize.asteroids:
                local_ast = True
            elif planet.size == fo.planetSize.gasGiant:
                local_gg = True
            spec_name = planet.speciesName
            this_spec = fo.getSpecies(spec_name)
            owner_id = planet.owner
            planet_population = planet.currentMeterValue(fo.meterType.population)
            buildings_here = [universe.getBuilding(bldg).buildingTypeName for bldg in planet.buildingIDs]
            weapons_grade = "WEAPONS_0.0"
            if owner_id == empire_id:
                if planet_population > 0.0 and this_spec:
                    empire_has_qualifying_planet = True
                    for metab in [tag for tag in this_spec.tags if tag in AIDependencies.metabolismBoostMap]:
                        empire_metabolisms[metab] = empire_metabolisms.get(metab, 0.0) + planet.habitableSize
                    if this_spec.canProduceShips:
                        pilot_val = rate_piloting(spec_name)
                        weapons_grade = "WEAPONS_%.1f" % pilot_val
                        set_pilot_rating_for_planet(pid, pilot_val)
                        yard_here = []
                        if "BLD_SHIPYARD_BASE" in buildings_here:
                            set_ship_builders(spec_name, pid)
                            yard_here = [pid]
                        if this_spec.canColonize and planet.currentMeterValue(fo.meterType.targetPopulation) >= 3:
                            set_colony_builders(spec_name, yard_here)
                set_building_locations(weapons_grade, buildings_here, pid)

                for special in planet.specials:
                    if special_is_nest(special):
                        set_have_nest()
                    if special in AIDependencies.metabolismBoosts:
                        set_growth_special(special, pid)
                        if planet.focus == FocusType.FOCUS_GROWTH:
                            active_growth_specials.setdefault(special, []).append(pid)
            elif owner_id != -1:
                if get_partial_visibility_turn(pid) >= current_turn - 1:  # only interested in immediately recent data
                    aistate.misc.setdefault("enemies_sighted", {}).setdefault(current_turn, []).append(pid)

        if empire_has_qualifying_planet:
            if local_ast:
                set_have_asteroids()
            elif local_gg:
                set_have_gas_giant()

        if sys_id in get_owned_planets():
            AIstate.empireStars.setdefault(system.starType, []).append(sys_id)
            sys_status = aistate.systemStatus.setdefault(sys_id, {})
            if sys_status.get("fleetThreat", 0) > 0:
                set_colonies_is_under_attack()
            if sys_status.get("neighborThreat", 0) > 0:
                set_colonies_is_under_treat()

    survey_universe_lock.unlock()
    survey_timer.stop()

    # lock must be released before this, since it uses the locked functions
    _print_empire_species_roster()


def _print_empire_species_roster():
    """Print empire species roster in table format to log."""
    grade_map = {
        "ULTIMATE": "+++",
        "GREAT": "++",
        "GOOD": "+",
        "AVERAGE": "o",
        "BAD": "-",
        "VERY_BAD": "--",
        "EXTREMELY_BAD": "---",
        "NO": "x",
    }

    grade_tags = [
        Tags.INDUSTRY,
        Tags.RESEARCH,
        Tags.INFLUENCE,
        Tags.POPULATION,
        Tags.SUPPLY,
        Tags.WEAPONS,
        Tags.ATTACKTROOPS,
    ]

    grade_tags_names = {
        Tags.INDUSTRY: "Ind.",
        Tags.RESEARCH: "Res.",
        Tags.INFLUENCE: "Infl.",
        Tags.POPULATION: "Pop.",
        Tags.SUPPLY: "Supply",
        Tags.WEAPONS: "Pilot",
        Tags.ATTACKTROOPS: "Troop",
    }

    species_table = Table(
        Text("species"),
        Sequence("PIDs"),
        Bool("Colonize"),
        Text("Shipyards"),
        *[Text(grade_tags_names[v]) for v in grade_tags],
        Sequence("Tags"),
        table_name="Empire species roster Turn %d" % fo.currentTurn(),
    )
    for species_name, planet_ids in get_empire_planets_by_species().items():
        species_tags = fo.getSpecies(species_name).tags
        number_of_shipyards = len(get_ship_builder_locations(species_name))
        species_table.add_row(
            species_name,
            planet_ids,
            can_build_colony_for_species(species_name),
            number_of_shipyards,
            *[grade_map.get(get_species_tag_grade(species_name, tag).upper(), "o") for tag in grade_tags],
            [tag for tag in species_tags if not any(s in tag for s in grade_tags) and "PEDIA" not in tag],
        )
    info(species_table)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_NDIM_SUBSPACE",
    description="LRN_NDIM_SUBSPACE_DESC",
    short_description="THEORY_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=128 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY", "THEORY"],
    prerequisites=["LRN_FORCE_FIELD"],
    graphic="icons/tech/n-dimensional_subspace.png",
)



from focs._effects import (
    AddSpecial,
    EffectsGroup,
    HasTag,
    IsSource,
    OwnedBy,
    OwnerHasTech,
    Planet,
    RemoveSpecial,
    Source,
    StatisticIf,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SPY_STEALTH_3",
    description="SPY_STEALTH_3_DESC",
    short_description="PLANET_PROTECT_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=(
        700
        * TECH_COST_MULTIPLIER
        / (1 + StatisticIf(float, condition=IsSource & OwnerHasTech(name="SPY_STEALTH_PART_3")))
        / (1 + StatisticIf(float, condition=OwnedBy(empire=Source.Owner) & HasTag(name="SNEAKY")))
    ),
    researchturns=7,
    tags=["PEDIA_SPY_CATEGORY"],
    prerequisites=["SPY_STEALTH_2"],
    effectsgroups=[
        EffectsGroup(
            scope=OwnedBy(empire=Source.Owner) & Planet(),
            activation=~OwnerHasTech(name="SPY_STEALTH_4"),
            effects=[
                AddSpecial(name="DIM_RIFT_SLAVE_SPECIAL"),
                RemoveSpecial(name="VOLCANIC_ASH_SLAVE_SPECIAL"),
                RemoveSpecial(name="CLOUD_COVER_SLAVE_SPECIAL"),
            ],
        )
    ],
    graphic="icons/specials_huge/dim_rift.png",
)



from focs._species import *
from species.species_macros.detection import COMMUNAL_VISION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, TUNDRA_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import GOOD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import GREAT_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_GEORGE",
    description="SP_GEORGE_DESC",
    gameplay_description="SP_GEORGE_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "TELEPATHIC",
        "GOOD_INDUSTRY",
        "BAD_RESEARCH",
        "GOOD_HAPPINESS",
        "GREAT_SUPPLY",
        "PEDIA_LITHIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "SUCCULENT_BARNACLES_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "MONOPOLE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "GAIA_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PANOPTICON_SPECIAL",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_CHARGE",
        "PLC_EXPLORATION",
        "PLC_NATIVE_APPROPRIATION",
        "PLC_THE_HUNT",
    ],
    dislikes=[
        "BLD_STARGATE",
        "BLD_GAS_GIANT_GEN",
        "BLD_SHIPYARD_CON_NANOROBO",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_CON_ADV_ENGINE",
        "BLD_GATEWAY_VOID",
        "BLD_ENCLAVE_VOID",
        "PHILOSOPHER_SPECIAL",
        "ELERIUM_SPECIAL",
        "FRUIT_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_CHECKPOINTS",
        "PLC_CONFORMANCE",
        "PLC_MODERATION",
        "PLC_NO_SUPPLY",
    ],
    effectsgroups=[
        *GOOD_INDUSTRY,
        *BAD_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_GEORGE"),
        *GREAT_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        BROAD_EP,
        *STANDARD_SHIP_SHIELDS,
        COMMUNAL_VISION("SP_GEORGE"),
    ],
    environments=TUNDRA_BROAD_EP,
    graphic="icons/species/george.png",
)



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    LocalCandidate,
    MinOf,
    NamedRealLookup,
    OwnedBy,
    Planet,
    SetDefense,
    SetMaxShield,
    SetShield,
    SetTroops,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import PLANET_DEFENSE_FACTOR, PLANET_SHIELD_FACTOR
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY
from techs.defense.mines import EG_SYSTEM_MINES

Tech(
    name="DEF_ROOT_DEFENSE",
    description="DEF_ROOT_DEFENSE_DESC",
    short_description="TROOPS_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            stackinggroup="PLANET_SHIELDS_STACK_ROOT",
            effects=SetMaxShield(value=Value + PLANET_SHIELD_FACTOR, accountinglabel="DEF_ROOT_DEFENSE"),
        ),
        EffectsGroup(  # base regeneration of troops, defense and shields if not attacked
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn)
            & (LocalCandidate.LastTurnConquered < CurrentTurn),
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=[
                SetShield(value=MinOf(float, Value + PLANET_SHIELD_FACTOR, Value(Target.MaxShield))),
                SetDefense(value=MinOf(float, Value + PLANET_DEFENSE_FACTOR, Value(Target.MaxDefense))),
                SetTroops(value=MinOf(float, Value + 1, Value(Target.MaxTroops))),
            ],
        ),
        EffectsGroup(  # set minimum troops for just-colonized planets
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnColonized == CurrentTurn),
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetTroops(
                value=MinOf(float, NamedRealLookup(name="IMPERIAL_GARRISON_MAX_TROOPS_FLAT"), Value(Target.MaxTroops))
            ),
        ),
    ],
    graphic="",
)

Tech(
    name="DEF_PLANET_CLOAK",
    description="DEF_PLANET_CLOAK_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=9999 * TECH_COST_MULTIPLIER,
    researchturns=9999,
    researchable=False,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["SPY_STEALTH_3"],
    unlock=Item(type=UnlockBuilding, name="BLD_PLANET_CLOAK"),
)

Tech(
    name="DEF_SYST_DEF_MINE_1",
    description="DEF_SYST_DEF_MINE_1_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=192 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_DEFENSE_NET_1"],
    effectsgroups=EG_SYSTEM_MINES(
        NamedRealLookup(name="DEF_SYST_DEF_MINE_1_DAMAGE"), 75, "EMPIRE"
    ),  # Priority deliberately not a macro and before all priority macros
    graphic="icons/tech/system_defense_mines.png",
)

Tech(
    name="DEF_SYST_DEF_MINE_2",
    description="DEF_SYST_DEF_MINE_2_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=360 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_SYST_DEF_MINE_1"],
    effectsgroups=EG_SYSTEM_MINES(NamedRealLookup(name="DEF_SYST_DEF_MINE_2_DAMAGE"), 65, "EMPIRE"),
    graphic="icons/tech/system_defense_mines.png",
)

Tech(
    name="DEF_SYST_DEF_MINE_3",
    description="DEF_SYST_DEF_MINE_3_DESC",
    short_description="DEFENSE_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_SYST_DEF_MINE_2"],
    effectsgroups=EG_SYSTEM_MINES(NamedRealLookup(name="DEF_SYST_DEF_MINE_3_DAMAGE"), 60, "EMPIRE"),
    graphic="icons/tech/system_defense_mines.png",
)



from focs._effects import (
    Abs,
    CurrentTurn,
    EffectsGroup,
    Focus,
    HasSpecies,
    Homeworld,
    IsSource,
    LocalCandidate,
    MinOf,
    OwnedBy,
    Planet,
    SetMaxStockpile,
    SetStockpile,
    Source,
    Target,
    TargetPopulation,
    Value,
)
from macros.base_prod import STOCKPILE_PER_POP
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
)


def _stockpile_per_pop_effectsgroup_snip(label: str, value) -> dict:
    return {
        "scope": IsSource,
        "activation": Planet(),
        "accountinglabel": label + "_STOCKPILE_LABEL",
        "priority": TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        "effects": SetMaxStockpile(value=value),
    }


def _stockpile_per_pop_effectsgroup(label: str, value):
    return EffectsGroup(description=label + "_STOCKPILE_DESC", **_stockpile_per_pop_effectsgroup_snip(label, value))


NO_STOCKPILE = []

_STANDARD_STOCKPILE = [
    EffectsGroup(  # increase or decrease towards target meter, when not recently conquered
        scope=IsSource,
        activation=Planet()
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.System.LastTurnBattleHere < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetStockpile(
            value=Value
            + MinOf(float, Abs(float, Value(Target.MaxStockpile) - Value), 1)
            * (1 - 2 * (Value > Value(Target.MaxStockpile)))
        ),
    ),
    # increase stockpile for species if Homeworld is set to stockpile focus
    EffectsGroup(
        scope=Planet()
        & OwnedBy(empire=Source.Owner)
        & HasSpecies(name=[Source.Species])
        & ~Homeworld(name=[Source.Species]),
        activation=Planet() & Focus(type=["FOCUS_STOCKPILE"]) & Homeworld(),
        stackinggroup="HOMEWORLD_STOCKPILE_FOCUS_BONUS_LABEL",
        accountinglabel="HOMEWORLD_STOCKPILE_FOCUS_BONUS_LABEL",
        effects=SetMaxStockpile(value=(Value + 2 * Target.Population * STOCKPILE_PER_POP)),
    ),
    # removes residual stockpile capacity from a dead planet
    EffectsGroup(scope=IsSource, activation=Planet() & TargetPopulation(high=0), effects=SetStockpile(value=0)),
]

BAD_STOCKPILE = [
    _stockpile_per_pop_effectsgroup("BAD", Value + 0.5 * Target.Population * STOCKPILE_PER_POP),
    *_STANDARD_STOCKPILE,
]


AVERAGE_STOCKPILE = [
    EffectsGroup(
        # Skip the AVERAGE_STOCKPILE_DESC, same as for the other *_STOCKPILE macros
        **_stockpile_per_pop_effectsgroup_snip("AVERAGE", Value + 1 * Target.Population * STOCKPILE_PER_POP)
    ),
    *_STANDARD_STOCKPILE,
]


GOOD_STOCKPILE = [
    _stockpile_per_pop_effectsgroup("GOOD", Value + 3 * Target.Population * STOCKPILE_PER_POP),
    *_STANDARD_STOCKPILE,
]


GREAT_STOCKPILE = [
    _stockpile_per_pop_effectsgroup("GREAT", Value + 10 * Target.Population * STOCKPILE_PER_POP),
    *_STANDARD_STOCKPILE,
]


ULTIMATE_STOCKPILE = [
    _stockpile_per_pop_effectsgroup("ULTIMATE", Value + 15 * Target.Population * STOCKPILE_PER_POP),
    *_STANDARD_STOCKPILE,
]



_basic_types = {str, int, float, bool}


def _is_new_type(val):
    """
    NewType is a function on Python3.9, in Python3.10 is changed to class.
    """
    return str(val).startswith("<function NewType.")


def get_name_for_mapping(annotation) -> str:
    if _is_new_type(annotation) or annotation in _basic_types:
        return annotation.__name__
    elif isinstance(annotation, str):
        return annotation
    else:
        args = [get_name_for_mapping(x) for x in annotation.__args__]
        return f"{annotation.__name__}[{', '.join(args)}]"



"""This module defines the decoding for the FreeOrion AI savegames.

The decoder is subclassed from the standard library json decoder and
uses its string parsing. However, the resulting objects will be interpreted
differently according to the encoding used in FreeOrion AI savegames.

The decoder will only load trusted classes as defined in _definitions.py,
if an unknown/untrusted object is encountered, it will raise a InvalidSaveGameException.

When classes are loaded, their __setstate__ method will be invoked if available or
the __dict__ content will be set directly. It is the responsiblity of the trusted classes
to provide a __setstate__ method to verify and possibly sanitize the content of the passed state.
"""
import binascii
import json
from typing import Union

import EnumsAI
from AIstate import AIstate

from ._definitions import (
    ENUM_PREFIX,
    FALSE,
    FLOAT_PREFIX,
    INT_PREFIX,
    NONE,
    PLACEHOLDER,
    SET_PREFIX,
    TRUE,
    TUPLE_PREFIX,
    InvalidSaveGameException,
    TrustedClasses,
)


class SaveDecompressException(Exception):
    """
    Exception class for troubles with decompressing save game string.
    """


def _starts_with_prefix(prefix: str, candidate: str) -> bool:
    return candidate.startswith(prefix)


def _extract_value(prefix: str, value: str):
    return value[len(prefix) :]


def _extract_collection(prefix: str, value: str):
    return value[len(prefix) + 1 : -1]


def load_savegame_string(string: Union[str, bytes]) -> AIstate:
    """
    :raises: SaveDecompressException, InvalidSaveGameException
    """
    import base64
    import zlib

    trusted_classes = TrustedClasses()
    try:
        new_string = base64.b64decode(string)
    except (binascii.Error, ValueError, TypeError) as e:
        raise SaveDecompressException("Fail to decode base64 savestate %s" % e) from e
    try:
        new_string = zlib.decompress(new_string)
    except zlib.error as e:
        raise SaveDecompressException("Fail to decompress savestate %s" % e) from e
    return decode(new_string.decode("utf-8"), trusted_classes)


def decode(obj, trusted_classes: TrustedClasses):
    return _FreeOrionAISaveGameDecoder(trusted_classes).decode(obj)


class _FreeOrionAISaveGameDecoder(json.JSONDecoder):
    def __init__(self, trusted_classes: TrustedClasses):
        # do not allow control characters
        self._trusted_classes = trusted_classes
        super().__init__(strict=True)

    def decode(self, s, _w=None):
        # use the default JSONDecoder to parse the string into a dict
        # then interpret the dict content according to our encoding
        retval = super().decode(s)
        return self.__interpret(retval)

    def __interpret_dict(self, obj):
        # if the dict does not contain the class-encoding keys,
        # then it is a standard dictionary.
        if not all(key in obj for key in ("__class__", "__module__")):
            return {self.__interpret(key): self.__interpret(value) for key, value in obj.items()}

        # pop and verify class and module name, then parse the class content
        class_name = obj.pop("__class__")
        module_name = obj.pop("__module__")
        full_name = f"{module_name}.{class_name}"
        cls = self._trusted_classes.get(full_name)
        if cls is None:
            raise InvalidSaveGameException("DANGER DANGER - %s not trusted" % full_name)

        parsed_content = self.__interpret_dict(obj)

        # create a new instance without calling the actual __new__or __init__
        # function of the class (so we avoid any side-effects from those)
        new_instance = object.__new__(cls)

        # Set the content trying to use the __setstate__ method if defined
        # Otherwise, directly set the dict content.
        try:
            setstate = new_instance.__setstate__
        except AttributeError:
            if not type(parsed_content) == dict:  # noqa: E721
                raise InvalidSaveGameException("Could not set content for %s" % new_instance)
            new_instance.__dict__ = parsed_content
        else:
            # only call now to not catch exceptions in the setstate method
            setstate(parsed_content)
        return new_instance

    def __interpret(self, x):  # noqa: C901
        """Interpret an object that was just decoded."""
        # primitive types do not have to be interpreted
        if isinstance(x, (int, float)):
            return x

        # special handling for dicts as they could encode our classes
        if isinstance(x, dict):
            return self.__interpret_dict(x)

        # for standard containers, interpret each element
        if isinstance(x, list):
            return list(self.__interpret(element) for element in x)

        # if it is a string, check if it encodes another data type
        if isinstance(x, str):
            # does it encode an integer?
            if _starts_with_prefix(INT_PREFIX, x):
                x = _extract_value(INT_PREFIX, x)
                return int(x)

            # does it encode a float?
            if _starts_with_prefix(FLOAT_PREFIX, x):
                x = _extract_value(FLOAT_PREFIX, x)
                return float(x)

            # does it encode a tuple?
            if _starts_with_prefix(TUPLE_PREFIX, x):
                # ignore surrounding parentheses
                content = _extract_collection(TUPLE_PREFIX, x)
                content = _replace_quote_placeholders(content)
                result = self.decode(content)
                return tuple(result)

            # does it encode a set?
            if _starts_with_prefix(SET_PREFIX, x):
                # ignore surrounding parentheses
                content = _extract_collection(SET_PREFIX, x)
                content = _replace_quote_placeholders(content)
                result = self.decode(content)
                return set(result)

            # does it encode an enum?
            if _starts_with_prefix(ENUM_PREFIX, x):
                full_name = _extract_value(ENUM_PREFIX, x)
                partial_names = full_name.split(".")
                if not len(partial_names) == 2:
                    raise InvalidSaveGameException("Could not decode Enum %s" % x)
                enum_name = partial_names[0]
                enum = getattr(EnumsAI, enum_name, None)
                if enum is None:
                    raise InvalidSaveGameException("Invalid enum %s" % enum_name)
                retval = getattr(enum, partial_names[1], None)
                if retval is None:
                    raise InvalidSaveGameException("Invalid enum value %s" % full_name)
                return retval

            if _starts_with_prefix(TRUE, x):
                return True

            if _starts_with_prefix(FALSE, x):
                return False

            if _starts_with_prefix(NONE, x):
                return None

            # no special cases apply at this point, should be a standard string
            return x

        raise TypeError(f"Unexpected type {type(x)} ({x})")


def _replace_quote_placeholders(s):
    """Replace PLACEHOLDER with quotes if not nested within another encoded container.

    To be able to use tuples as dictionary keys, to use standard json decoder,
    the entire tuple with its content must be encoded as a single string.
    The inner objects may no longer be quoted as that would prematurely terminate
    the strings. Inner quotes are therefore replaced with the PLACEHOLDER char.

    Example:
        output = encode(tuple(["1", "string"]))
        "__TUPLE__([$1$, $string$])"

    To be able to decode the inner content, the PLACEHOLDER must be converted
    to quotes again.
    """
    n = 0  # counts nesting level (i.e. number of opened but not closed parentheses)
    start = 0  # starting point for string replacement
    for i in range(len(s)):
        if s[i] == "(":
            # if this is an outer opening parenthesis, then replace placeholder from last parenthesis to here
            if n == 0:
                s = s[:start] + s[start:i].replace(PLACEHOLDER, '"') + s[i:]
            n += 1
        elif s[i] == ")":
            n -= 1
            if n == 0:
                start = i
    s = s[:start] + s[start:].replace(PLACEHOLDER, '"')
    return s



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_UNIF_CONC",
    description="LRN_UNIF_CONC_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=630 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_GATEWAY_VOID", "LRN_QUANT_NET"],
    unlock=[Item(type=UnlockBuilding, name="BLD_COLLECTIVE_NET"), Item(type=UnlockPolicy, name="PLC_DREAM_RECURSION")],
    graphic="icons/tech/unified_conscious.png",
)



"""
Various actions for the order generation.
"""
import freeOrionAIInterface as fo
import random
from functools import wraps
from logging import debug, error, fatal, info

import PlanetUtilsAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate

# This import initializes the aggresion module and we need to import it to make module work
from character.character_module import Aggression  # noqa F401
from character.character_strings_module import get_trait_name_aggression
from common.option_tools import check_bool, get_option_dict
from common.print_utils import dict_to_table
from freeorion_tools.statistics import stats


def _once(function):
    """
    Run once per runtime.
    """
    executed = False

    @wraps(function)
    def wrapper():
        nonlocal executed
        if executed:
            return
        else:
            executed = True
            function()

    return wrapper


@_once
def print_existing_rules():
    try:
        rules = fo.getGameRules()
        debug(dict_to_table(rules.getRulesAsStrings(), name="Defined game rules"))
    except Exception as e:
        error("Exception %s when trying to get game rules" % e, exc_info=True)


def empire_is_ok():
    """
    If nothing can be ordered anyway, exit early.
    Note that there is no need to update meters etc. in this case.

    Return False if we should stop order generation.
    """
    empire = fo.getEmpire()
    if empire is None:
        fatal("This client has no empire. Aborting order generation.")
        return False
    elif empire.eliminated:
        info("This empire has been eliminated. Aborting order generation.")
        return False
    return True


def update_resource_pool():
    info("Meter / Resource Pool updating...")
    fo.initMeterEstimatesDiscrepancies()
    fo.updateMeterEstimates(False)
    fo.updateResourcePools()


def set_game_turn_seed():
    """
    Set the random seed for game-reload consistency.

    Seed is based on galaxy seed, empire name and current turn.
    """
    turn = fo.currentTurn()
    random_seed = str(fo.getGalaxySetupData().seed) + "%05d%s" % (turn, fo.getEmpire().name)
    random.seed(random_seed)


def _dump_empire_info():
    empire = fo.getEmpire()
    turn = fo.currentTurn()

    # TODO: It would be nice to decouple char and AI state.
    #       Character is immutable and ai state is mutable.
    #       Dependency on immutable object is easier to manage
    research_index = get_aistate().character.get_research_index()
    aggression_name = get_trait_name_aggression(get_aistate().character)

    name_parts = (
        empire.name,
        empire.empireID,
        "pid",
        fo.playerID(),
        fo.playerName(),
        "RIdx",
        research_index,
        aggression_name.capitalize(),
    )
    empire_name = "_".join(str(part) for part in name_parts)
    stats.empire(empire.empireID, empire_name, turn)
    stats.empire_color(*empire.colour)


def greet_on_first_turn(diplomatic_corp):
    if fo.currentTurn() != 1:
        return

    human_player = fo.empirePlayerID(1)
    greet = diplomatic_corp.get_first_turn_greet_message()
    trait_name = get_trait_name_aggression(get_aistate().character)
    fo.sendChatMessage(human_player, f"{fo.getEmpire().name} ({trait_name}): [[{greet}]]")


def replay_turn_after_load():
    """
    When loading a savegame, the AI will already have issued orders for this turn.
    To avoid duplicate orders, generally try not to replay turns. However, for debugging
    purposes it is often useful to replay the turn and observe varying results after
    code changes. Set the replay_after_load flag in the AI config to let the AI issue
    new orders after a game load. Note that the orders from the original savegame are
    still being issued and the AIstate was saved after those orders were issued.

    Return True is we should abort order generation.
    """
    replay_option_is_set = check_bool(get_option_dict().get("replay_turn_after_load", "False"))
    turn_is_already_played = fo.currentTurn() == get_aistate().last_turn_played

    # TODO: Consider adding an option to clear AI orders after load (must save AIstate at turn start then)
    if turn_is_already_played:
        info("The AIstate indicates that this turn was already played.")
        if replay_option_is_set:
            info("Issuing new orders anyway.")
            return False
        else:
            info("Aborting new order generation. Orders from savegame will still be issued.")
            return True
    else:
        return False


def print_starting_intro():
    debug("\n\n\n" + "=" * 20)
    debug(f"Starting turn {fo.currentTurn()}")
    debug("=" * 20 + "\n")

    debug("***************************************************************************")
    debug("*******  Log info for AI progress chart script. Do not modify.   **********")
    debug("Generating Orders")
    _dump_empire_info()
    _print_empire_capital()
    stats.ship_count(get_aistate().shipCount)


def _print_empire_capital():
    planet_id = PlanetUtilsAI.get_capital()
    if planet_id is not None and planet_id != INVALID_ID:
        planet = fo.getUniverse().getPlanet(planet_id)
        stats.capital(planet_id, planet.name, planet.speciesName)
    else:
        stats.capital(None, None, None)
    debug("***************************************************************************")
    debug("***************************************************************************")



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_TERRAFORM",
    description="GRO_TERRAFORM_DESC",
    short_description="GRO_TERRAFORM_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=160 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_ADV_ECOMAN"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_BEST"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_TERRAN"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_OCEAN"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_SWAMP"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_TOXIC"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_INFERNO"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_RADIATED"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_BARREN"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_TUNDRA"),
        Item(type=UnlockBuilding, name="BLD_TERRAFORM_DESERT"),
        Item(type=UnlockPolicy, name="PLC_ENVIRONMENTALISM"),
    ],
    graphic="icons/tech/terraform.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_GENETIC_MED",
    description="GRO_GENETIC_MED_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=24 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_GENETIC_ENG"],
    unlock=Item(type=UnlockBuilding, name="BLD_GENOME_BANK"),
    graphic="icons/tech/genetic_medicine.png",
)



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    LocalCandidate,
    MaxOf,
    MinOf,
    NamedReal,
    OwnedBy,
    Planet,
    SetMaxShield,
    SetShield,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import PLANET_SHIELD_FACTOR
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY, DEFAULT_PRIORITY

Tech(
    name="DEF_PLAN_BARRIER_SHLD_1",
    description="DEF_PLAN_BARRIER_SHLD_1_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=125 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["LRN_FORCE_FIELD"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value + NamedReal(name="DEF_PLAN_BARRIER_SHLD_1_MAX_SHIELD_FLAT", value=30 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn - 1),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetShield(
                value=Value
                + MaxOf(
                    float, MinOf(float, PLANET_SHIELD_FACTOR, Value), 0.25 * Target.Construction * PLANET_SHIELD_FACTOR
                )
            ),
        ),
    ],
    graphic="icons/tech/planetary_barrier_shield.png",
)

Tech(
    name="DEF_PLAN_BARRIER_SHLD_2",
    description="DEF_PLAN_BARRIER_SHLD_2_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=192 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_PLAN_BARRIER_SHLD_1"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value + NamedReal(name="DEF_PLAN_BARRIER_SHLD_2_MAX_SHIELD_FLAT", value=60 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn - 1),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetShield(
                value=Value
                + MaxOf(
                    float,
                    MinOf(float, 3.0 * PLANET_SHIELD_FACTOR, Value),
                    0.75 * Target.Construction * PLANET_SHIELD_FACTOR,
                )
            ),
        ),
    ],
    graphic="icons/tech/planetary_barrier_shield.png",
)

Tech(
    name="DEF_PLAN_BARRIER_SHLD_3",
    description="DEF_PLAN_BARRIER_SHLD_3_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=360 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_PLAN_BARRIER_SHLD_2"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value + NamedReal(name="DEF_PLAN_BARRIER_SHLD_3_MAX_SHIELD_FLAT", value=90 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn - 1),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetShield(
                value=Value
                + MaxOf(
                    float,
                    MinOf(float, 5.0 * PLANET_SHIELD_FACTOR, Value),
                    1.0 * Target.Construction * PLANET_SHIELD_FACTOR,
                )
            ),
        ),
    ],
    graphic="icons/tech/planetary_barrier_shield.png",
)

Tech(
    name="DEF_PLAN_BARRIER_SHLD_4",
    description="DEF_PLAN_BARRIER_SHLD_4_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_PLAN_BARRIER_SHLD_3"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value
                + NamedReal(name="DEF_PLAN_BARRIER_SHLD_4_MAX_SHIELD_FLAT", value=150 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn - 1),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetShield(
                value=Value
                + MaxOf(
                    float,
                    MinOf(float, 9.0 * PLANET_SHIELD_FACTOR, Value),
                    1.5 * Target.Construction * PLANET_SHIELD_FACTOR,
                )
            ),
        ),
    ],
    graphic="icons/tech/planetary_barrier_shield.png",
)

Tech(
    name="DEF_PLAN_BARRIER_SHLD_5",
    description="DEF_PLAN_BARRIER_SHLD_5_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=1200 * TECH_COST_MULTIPLIER,
    researchturns=12,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_PLAN_BARRIER_SHLD_4"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value
                + NamedReal(name="DEF_PLAN_BARRIER_SHLD_5_MAX_SHIELD_FLAT", value=150 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn - 1),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetShield(
                value=Value
                + MaxOf(
                    float,
                    MinOf(float, 14.0 * PLANET_SHIELD_FACTOR, Value),
                    2.5 * Target.Construction * PLANET_SHIELD_FACTOR,
                )
            ),
        ),
    ],
    graphic="icons/tech/planetary_barrier_shield.png",
)



import freeOrionAIInterface as fo

import AIDependencies
from common.fo_typing import PlanetId
from EnumsAI import FocusType
from freeorion_tools.caching import cache_for_current_turn
from turn_state._planet_state import _get_planets_info


class EmpireResources:
    def __init__(self):  # noqa: C901
        self.have_gas_giant = False
        self.have_asteroids = False
        self.owned_asteroid_coatings = 0
        self.have_ruins = False
        self.have_nest = False
        self.have_computronium = False
        self.pids_computronium = []
        self.have_honeycomb = False
        self.pids_honeycomb = []
        self.have_worldtree = False
        self.num_researchers = 0  # population with research focus
        self.num_industrialists = 0  # population with industry focus
        self.luxury_planets = {}  # list of planets per luxury special

        empire_id = fo.empireID()
        universe = fo.getUniverse()
        INDUSTRY = FocusType.FOCUS_INDUSTRY
        RESEARCH = FocusType.FOCUS_RESEARCH

        for planet_info in _get_planets_info().values():
            planet = universe.getPlanet(planet_info.pid)
            if planet.ownedBy(empire_id):
                if AIDependencies.ANCIENT_RUINS_SPECIAL in planet.specials:
                    self.have_ruins = True

                if AIDependencies.WORLDTREE_SPECIAL in planet.specials:
                    self.have_world_tree = True

                if AIDependencies.ASTEROID_COATING_OWNED_SPECIAL in planet.specials:
                    self.owned_asteroid_coatings += 1

                if AIDependencies.COMPUTRONIUM_SPECIAL in planet.specials and RESEARCH in planet.availableFoci:
                    self.pids_computronium.append(planet.id)
                    if planet.focus == RESEARCH:
                        self.have_computronium = True

                if AIDependencies.HONEYCOMB_SPECIAL in planet.specials and INDUSTRY in planet.availableFoci:
                    self.pids_honeycomb.append(planet.id)
                    if planet.focus == INDUSTRY:
                        self.have_honeycomb = True

                for special in AIDependencies.luxury_specials:
                    if special in planet.specials:
                        self.luxury_planets.setdefault(special, []).append(planet)

                population = planet.currentMeterValue(fo.meterType.population)
                if planet.focus == INDUSTRY:
                    self.num_industrialists += population
                elif planet.focus == RESEARCH:
                    self.num_researchers += population


@cache_for_current_turn
def _get_planet_catalog() -> EmpireResources:
    return EmpireResources()


def population_with_industry_focus() -> int:
    return _get_planet_catalog().num_industrialists


def have_gas_giant() -> bool:
    return _get_planet_catalog().have_gas_giant


def set_have_gas_giant():
    _get_planet_catalog().have_gas_giant = True


def have_asteroids() -> bool:
    return _get_planet_catalog().have_asteroids


def owned_asteroid_coatings() -> int:
    return _get_planet_catalog().owned_asteroid_coatings


def set_have_asteroids():
    _get_planet_catalog().__have_asteroids = True


def have_ruins() -> bool:
    return _get_planet_catalog().have_ruins


def have_nest() -> bool:
    return _get_planet_catalog().have_nest


def set_have_nest():
    _get_planet_catalog().have_nest = True


def have_computronium() -> bool:
    """Return True if we have a planet with a computronium moon, which is set to research focus."""
    return _get_planet_catalog().have_computronium


def computronium_candidates() -> list[PlanetId]:
    """Return list of own planets that have a computronium moon and a species capable of research."""
    return _get_planet_catalog().pids_computronium


def have_honeycomb() -> bool:
    """Return True if we have a planet with a honeycomb special, which is set to production focus."""
    return _get_planet_catalog().have_honeycomb


def honeycomb_candidates() -> list[PlanetId]:
    """Return list of own planets that have the honeycomb special and a species capable of production."""
    return _get_planet_catalog().pids_honeycomb


def have_worldtree() -> bool:
    return _get_planet_catalog().have_worldtree


def luxury_resources() -> dict[str, list[fo.planet]]:
    return _get_planet_catalog().luxury_planets



from collections import defaultdict
from enum import Enum
from logging import warning

from stub_generator.interface_inspector.class_processor import ClassInfo, inspect_class
from stub_generator.interface_inspector.enum_processor import EnumInfo, inspect_enum
from stub_generator.interface_inspector.function_processor import (
    FunctionInfo,
    inspect_function,
)
from stub_generator.interface_inspector.inspection_helpers import _getmembers
from stub_generator.interface_inspector.instance_processing import (
    InstanceInfo,
    inspect_instances,
)


class MemberType(Enum):
    ENUM = 1
    CLASS = 2
    FUNCTION = 3
    UNKNOWN = -1


def get_type(member):
    type_ = str(type(member))
    if type_ == "<class 'type'>":
        return MemberType.ENUM
    elif type_ == "<class 'Boost.Python.class'>":
        return MemberType.CLASS
    elif type_ == "<class 'Boost.Python.function'>":
        return MemberType.FUNCTION
    else:
        return MemberType.UNKNOWN


_OBJECT_HANDLERS = {
    MemberType.ENUM: inspect_enum,
    MemberType.CLASS: inspect_class,
    MemberType.FUNCTION: inspect_function,
}

_MODULE_NAMES_TO_IGNORE = {
    "__doc__",
    "__package__",
    "__name__",
    "INVALID_GAME_TURN",
    "to_str",
    "__loader__",
    "__spec__",
}


def get_module_members(obj):
    module_members = defaultdict(list)

    for name, member in _getmembers(obj):
        if name in _MODULE_NAMES_TO_IGNORE:
            continue
        type_key = get_type(member)
        if type_key == MemberType.UNKNOWN:
            warning(f"Unknown: '{name}' of type '{type(member)}': {member}")
        else:
            type_inspector = _OBJECT_HANDLERS[type_key]
            module_members[type_key].append(type_inspector(name, member))
    return module_members


def get_module_info(obj, instances) -> tuple[list[ClassInfo], list[EnumInfo], list[FunctionInfo], list[InstanceInfo]]:
    module_members = get_module_members(obj)
    return (
        module_members[MemberType.CLASS],
        module_members[MemberType.ENUM],
        module_members[MemberType.FUNCTION],
        list(inspect_instances(instances)),
    )



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SPACE_FLUX_DRIVE",
    description="SHP_SPACE_FLUX_DRIVE_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=40 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_FLUX_HULL_TECHS"],
    prerequisites=["SHP_SPACE_FLUX_BASICS"],
    unlock=Item(type=UnlockShipHull, name="SH_SPATIAL_FLUX"),
    graphic="icons/ship_hulls/spatial_flux_hull_small.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ZERO_POINT",
    description="SHP_ZERO_POINT_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=270 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_FUEL_PART_TECHS"],
    prerequisites=["SHP_ANTIMATTER_TANK", "PRO_ZERO_GEN"],
    unlock=Item(type=UnlockShipPart, name="FU_ZERO_FUEL"),
    graphic="icons/ship_parts/zero-point-generator.png",
)



from focs._effects import Blue, EffectsGroup, InSystem, IsSource, SetFuel, Ship, Star, Stationary, Value, White
from focs._species import *
from species.species_macros.detection import GREAT_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, OCEAN_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import INDEPENDENT_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import GOOD_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import GOOD_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS

Species(
    name="SP_LAENFA",
    description="SP_LAENFA_DESC",
    gameplay_description="SP_LAENFA_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "PHOTOTROPHIC",
        "SNEAKY",
        "TELEPATHIC",
        "BAD_INFLUENCE",
        "GOOD_POPULATION",
        "AVERAGE_SUPPLY",
        "GREAT_DETECTION",
        "GOOD_STEALTH",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_PC_FUEL",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
        "INDEPENDENT_HAPPINESS",
        "PEDIA_TELEPATHIC_TITLE",
        "PEDIA_SNEAKY_TITLE",
        "PEDIA_INDEPENDENT",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "FOCUS_STOCKPILE",
        "FRUIT_SPECIAL",
        "PANOPTICON_SPECIAL",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_REGIONAL_ADMIN",
        "PLC_BUREAUCRACY",
        "PLC_CENTRALIZATION",
        "PLC_CONFORMANCE",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_DIVINE_AUTHORITY",
        "PLC_DREAM_RECURSION",
        "PLC_ENVIRONMENTALISM",
        "PLC_EXPLORATION",
        "PLC_FLANKING",
        "PLC_INSURGENCY",
        "PLC_NATIVE_APPROPRIATION",
        "POSITRONIUM_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "SP_CELESTEPHYTE",  # because they too are telepathic plants
        "SUCCULENT_BARNACLES_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
    ],
    dislikes=[
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "BLD_BIOTERROR_PROJECTOR",
        "BLD_CULTURE_LIBRARY",
        "BLD_GAS_GIANT_GEN",
        "BLD_INDUSTRY_CENTER",
        "BLD_PLANET_BEACON",
        "BLD_SOL_ORB_GEN",
        "ELERIUM_SPECIAL",
        "PLC_EXOBOT_PRODUCTIVITY",
        "PLC_INDUSTRIALISM",
        "PLC_MARINE_RECRUITMENT",
        "RESONANT_MOON_SPECIAL",
        "WORLDTREE_SPECIAL",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *BAD_INFLUENCE,
        *GOOD_STOCKPILE,
        *GOOD_POPULATION,
        *INDEPENDENT_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_LAENFA"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *GREAT_DETECTION,
        *GOOD_STEALTH,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        BROAD_EP,
        *STANDARD_SHIP_SHIELDS,
        # replenish fuel in suitably coloured star systems,
        EffectsGroup(
            description="PHOTOTROP_DESC",
            scope=IsSource,
            activation=Ship & Stationary & InSystem() & Star(type=[Blue]),
            effects=SetFuel(value=Value + 1.9),  # sums with natural 0.1 per turn to a nice round 2.0 per turn
        ),
        EffectsGroup(
            scope=IsSource,
            activation=Ship & Stationary & InSystem() & Star(type=[White]),
            effects=SetFuel(value=Value + 0.9),
        ),
    ],
    environments=OCEAN_BROAD_EP,
    graphic="icons/species/laenfa.png",
)



from planets import planet_sizes, planet_types

#############################
# STAR GROUP NAMING OPTIONS #
#############################

# if star_groups_use_chars is true use entries from the stringtable entry STAR_GROUP_CHARS (single characters),
# otherwise use words like 'Alpha' from the stringtable entry STAR_GROUP_WORDS.
STAR_GROUPS_USE_CHARS = True

# postfix_stargroup_modifiers determines how star group system names are modified.
# if true, the greek letters (as chars or words, determined by star_group_use_chars,
# see above) are appended as postfix ("Centauri Alpha"), if false, they are prepended
# as prefix ("Alpha Centauri").
POSTFIX_STARGROUP_MODIFIERS = True

# the target proportion of star systems to be given individual names, dependent on size of galaxy
TARGET_INDIV_RATIO_SMALL = 0.6  # in small galaxies, 60% of the star systems get individual names
TARGET_INDIV_RATIO_LARGE = 0.3  # in large galaxies, 30% of the star systems get individual names
NAMING_LARGE_GALAXY_SIZE = 200  # a galaxy with 200+ star systems is considered as large, below that value as small


#################################
# HOME SYSTEM SELECTION OPTIONS #
#################################

# These options are needed in the home system selection/placement process. They determine the minimum number of
# systems and planets that a home system must have in its near vicinity, define the extend of this "near vicinity", etc.

# The following two options are used to determine the minimum number of systems and planets. This limit is
# HS_MIN_SYSTEMS_IN_VICINITY systems and HS_MIN_PLANETS_IN_VICINITY_PER_SYSTEM planets per system within the near
# vicinity of a home system, capped at HS_MIN_PLANETS_IN_VICINITY_TOTAL
HS_MIN_SYSTEMS_IN_VICINITY = 8
HS_MIN_PLANETS_IN_VICINITY_TOTAL = 10
HS_MIN_PLANETS_IN_VICINITY_PER_SYSTEM = 1

# The following options is used to determine the minimum spread of planet types in the vicinity of the home systems.
# Testing suggests around 8 planets is the approximate median without this check in place (based on a 320 system disc
# with 7 players, and all average settings).
HS_MIN_PLANET_TYPES_IN_VICINITY = 8

# This option defines the extend of what is considered the "near vicinity" of a home system. This are all systems that
# are within the number of jumps specified by HS_VICINITY_RANGE.
HS_VICINITY_RANGE = 3

# This options sets the maximum starting value for the minimum jump distance limit required between home systems.
# With large galaxies an excessive amount of time can be used in failed attempts to select home systems, so defining
# an upper limit for the home system selection process to use when calculating the starting value for the minimum
# jump distance limit is reasonable.
HS_MAX_JUMP_DISTANCE_LIMIT = 10

# This options defines the minimum jump distance limit between home systems that should be considered high priority.
# As long as the jump distance limit which home systems must at least be apart does not get reduced below this limit
# during the home system selection process, the minimum systems in home system vicinity requirement takes
# precedence over the jump distance limit. If the jump distance limit drops below this minimum jump distance limit,
# the process is restarted giving the jump distance limit precedence.
HS_MIN_DISTANCE_PRIORITY_LIMIT = 5

# These two options define which types of planets are counted when determining the number of planets in the near
# vicinity of a home system. HS_ACCEPTABLE_PLANET_SIZES is actually only needed for the process of adding planets
# to the near vicinity of a home system in case that's needed to meet the limit.
HS_ACCEPTABLE_PLANET_TYPES = planet_types
HS_ACCEPTABLE_PLANET_SIZES = planet_sizes



from collections import OrderedDict

import pytest

from ColonisationAI import OrbitalColonizationManager
from EnumsAI import MissionType
from savegame_codec import CanNotSaveGameException, load_savegame_string
from savegame_codec._decoder import SaveDecompressException


@pytest.fixture()
def patch_orb_col_management(monkeypatch):
    def eq(self, other):
        self_tuple = self._colonization_plans, self.num_enqueued_bases
        other_tuple = other._colonization_plans, other.num_enqueued_bases
        return self_tuple == other_tuple

    monkeypatch.setattr(OrbitalColonizationManager, "__eq__", eq)


@pytest.fixture(
    params=[
        True,
        False,
        None,
        1,
        1.0,
        tuple(),
        tuple("abc"),
        ("1", ("2", ("3",))),
        "hello 'world'",
        [],
        [1, 2, 3],
        {},
        {1: 2, 3: 4},
        OrderedDict({}),
        OrderedDict({1: 2}),
        set(),
        {1, 2, 3},
        MissionType.COLONISATION,
        OrbitalColonizationManager(),
        [[[OrbitalColonizationManager(), MissionType.COLONISATION, 1], "2", 3.0, {1: 2}]],
        {1: {2: {3: OrbitalColonizationManager()}}},
    ]
)
def serialized_object(request, pack_object):
    obj = request.param
    yield obj, pack_object(obj)


def test_inherited_trusted_object_is_not_encoded(pack_object):
    class Inherited(OrbitalColonizationManager):
        pass

    with pytest.raises(CanNotSaveGameException, match="Class .*.Inherited is not trusted"):
        pack_object(Inherited())


@pytest.mark.parametrize("obj", ['"', "\\", "\n", "\t"])
def test_string_with_sepcial_caracters_encoded_and_decoded_back(obj, pack_object):
    encoded = pack_object(obj)
    assert obj == load_savegame_string(encoded)


@pytest.mark.usefixtures("patch_orb_col_management")
def test_string_decode_to_object(serialized_object):
    obj, serialized_string = serialized_object
    assert obj == load_savegame_string(serialized_string)


@pytest.mark.usefixtures("patch_orb_col_management")
def test_bytes_decode_to_object(serialized_object):
    obj, serialized_bytes = serialized_object
    assert obj == load_savegame_string(serialized_bytes.encode("utf-8"))


def test_not_base64_decode_raises_error():
    with pytest.raises(SaveDecompressException, match="Fail to decode base64 savestate"):
        load_savegame_string("///")


def test_not_gziped_decode_raises_error():
    with pytest.raises(SaveDecompressException, match="Fail to decompress savestate"):
        load_savegame_string("aaaa")


def test_not_unicode_decode_raises_error():
    with pytest.raises(SaveDecompressException, match="Fail to decode base64 savestate"):
        load_savegame_string("шишка")



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_ADV_ECOMAN",
    description="GRO_ADV_ECOMAN_DESC",
    short_description="THEORY_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=78 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_GROWTH_CATEGORY", "THEORY"],
    prerequisites=["GRO_SUBTER_HAB"],
    graphic="icons/tech/advanced_eco_man.png",
)



# Only check for own buildings. The client may have seen a building once, which has been
# destroyed while outside vision range. In this case the building remains forever in the
# client's context, but it should not stop a player from rebuilding the same type.
from focs._effects import BuildBuilding, Contains, CurrentContent, Enqueued, IsBuilding, OwnedBy, Source

ENQUEUE_BUILD_ONE_PER_PLANET = (
    ~Contains(IsBuilding(name=[CurrentContent]) & OwnedBy(empire=Source.Owner))
    & ~Enqueued(type=BuildBuilding, name=CurrentContent)
    & OwnedBy(empire=Source.Owner)
)

ENQUEUE_ARTIFICIAL_PLANET_EXCLUSION = (
    ~Contains(IsBuilding(name=["BLD_ART_PLANET"]))
    & ~Contains(IsBuilding(name=["BLD_ART_FACTORY_PLANET"]))
    & ~Contains(IsBuilding(name=["BLD_ART_PARADISE_PLANET"]))
    & ~Enqueued(type=BuildBuilding, name="BLD_ART_PLANET")
    & ~Enqueued(type=BuildBuilding, name="BLD_ART_FACTORY_PLANET")
    & ~Enqueued(type=BuildBuilding, name="BLD_ART_PARADISE_PLANET")
)



"""configure_logging redirects print and the logger to use the freeorion server.

Output redirected to the freeorion server prints in the appropriate log:
freeorion.log freeoriond.log or AI_<N>.log.

logging messages of levels warning and above are decorated with module name, file name, function name and line number.

Usage:

from common.configure_logging import redirect_logging_to_freeorion_logger
redirect_logging_to_freeorion_logger(log_level)

Then use python logging or print and have it re-directed appropriately.

Notes on using the standard python logger: The python standard library is
composed of two basic parts: loggers and handlers.  Loggers generate the log and
associate level, timing, filename, function, line number and other information
with the log.  Handlers convert the logging stream to file, console, email etc.

* Using the root logger directly is the simplest way to log:
logging.debug(msg)
logging.info(msg)
logging.warning(msg)
logging.error(msg)
logging.fatal(msg)

* The log level of the root logger can be changed with:

logging.getLogger().setLevel(logging.WARN)

* The log level of the root logger can be checked with:

logging.getLogger().isEnableFor(logging.WARN)


* One strength of the python standard logging library is the ability to create
  arbitrary hierarchical loggers.  Loggers are created globally when getLogger()
  is called with any string.  The hierarchical levels are dot separated.  The
  root logger is the empty string.  The following creates a logger:

logging.getLogger("toplevel.2ndlevel")

* Loggers can be filtered by log level hierarchically.  The following turns off
  logging below warning level for all loggers and then turns debug logging on
  for "name.subname" and its children.  This means that the log stream only
  contains logging for that one component.

logging.getLogger().setLevel(logging.ERROR)
logging.getLogger(name.subname).setLevel(logging.DEBUG)

* Logger formats string using the old format style.  Importantly, it only
  formats the string if that level of debugging is enabled.  This means that if
  the formatting is left to the logging sub-system, then there is no overhead to
  format the string if that log level is not enabled.

logging("A formatted %s contained the number nine %d", "string", 9)


* The following log levels have sinks in the game engine that write to the
  logfiles and include their level.  They should be used as follows:

FATAL   - used for errors that will crash the python engine.
ERROR   - used of "major" unrecoverable errors which will affect game play.
          For example being unable to place all the homeworlds.
WARNING - used for "minor", recoverable errors.
          For example not getting a ship object from a (supposedly valid) ship id.
INFO    - used to report game state and progress.
          For example showing a table about colonization possibilities or threat assessment.
DEBUG   - used for low-level implementation or calculation details.
          For example the calculations when rating a fleet.

For more information about the python standard library logger see
https://docs.python.org/2/howto/logging.html
"""
import logging
import os
import sys
import traceback

try:
    import freeorion_logger  # FreeOrion logger interface pylint: disable=import-error
except ImportError:
    # Create an alternative logger for use in testing when the server is unavailable
    class _FreeOrionLoggerForTest:
        """A stub freeorion_logger for testing"""

        @staticmethod
        def debug(msg, *args):
            print(msg)

        @staticmethod
        def info(msg, *args):
            print(msg)

        @staticmethod
        def warn(msg, *args):
            print(msg)

        @staticmethod
        def error(msg, *args):
            print(msg, file=sys.stderr)

        @staticmethod
        def fatal(msg, *args):
            print(msg, file=sys.stderr)

    freeorion_logger = _FreeOrionLoggerForTest()


class _stdXLikeStream:
    """A stream-like object to redirect stdout or stderr to the C++ process."""

    def __init__(self, level):
        self.logger = {
            logging.DEBUG: freeorion_logger.debug,
            logging.INFO: freeorion_logger.info,
            logging.WARNING: freeorion_logger.warn,
            logging.ERROR: freeorion_logger.error,
            logging.FATAL: freeorion_logger.fatal,
            logging.NOTSET: freeorion_logger.debug,
        }[level]

    def write(self, msg):
        # Grab the caller's call frame info
        if hasattr(sys, "_getframe"):
            frame = sys._getframe(1)
            try:
                line_number = frame.f_lineno
                filename = frame.f_code.co_filename
            except:  # noqa: E722
                (filename, line_number) = ("", "")
            finally:
                # Explicitly del references to the caller's frame to avoid persistent reference cycles
                del frame
        else:
            (filename, line_number) = ("", "")

        try:
            self.logger(msg, str(os.path.split(filename)[1]), line_number)

        finally:
            # Explicitly del references to the caller's frame to avoid persistent reference cycles
            del filename
            del line_number


class _LoggerHandler(logging.Handler):
    """A handler to send logs to the C++ process."""

    def __init__(self, level):
        super().__init__(level)
        self.logger = {
            logging.DEBUG: freeorion_logger.debug,
            logging.INFO: freeorion_logger.info,
            logging.WARNING: freeorion_logger.warn,
            logging.ERROR: freeorion_logger.error,
            logging.FATAL: freeorion_logger.fatal,
            logging.NOTSET: freeorion_logger.debug,
        }[level]

    def emit(self, record):
        msg = self.format(record)
        # By default, there is no line terminator in the msg
        # and the C++ backend will buffer the input until
        # newline is reached, i.e. will not add any terminator itself.
        # Since we don't want to manually add a newline terminator in
        # each logging call, ensure proper line termination.
        # Note that in the logging library, Handler.emit() is responsible
        # to add the required terminator, not Handler.format()
        if not msg.endswith("\n"):
            msg += "\n"
        self.logger(msg, record.filename, record.lineno)


class FOLogFormatter(logging.Formatter):
    def formatException(self, ei):
        exc_text = "".join(traceback.format_exception(*ei))
        tb = ei[2]
        if not tb:
            return exc_text
        while tb.tb_next:
            tb = tb.tb_next

        exc_text += "Local context: "
        exc_text += str(tb.tb_frame.f_locals)
        exc_text += "\n"
        return exc_text


class _SingleLevelFilter(logging.Filter):
    """This filter selects for only one log level."""

    def __init__(self, _level):
        super().__init__()
        self.level = _level

    def filter(self, record):
        return record.levelno == self.level


def _create_narrow_handler(level):
    """Create a handler for logger that forwards a single level of log
    to the appropriate stream in the C++ app."""
    h = _LoggerHandler(level)
    h.setFormatter(FOLogFormatter())
    h.addFilter(_SingleLevelFilter(level))
    h.setLevel(level)
    return h


def _unhandled_exception_hook(*exc_info):
    traceback_msg = "Uncaught exception: {}".format("".join(traceback.format_exception(*exc_info)))
    logging.getLogger().error(traceback_msg)


def redirect_logging_to_freeorion_logger(initial_log_level=logging.DEBUG):
    """Redirect stdout, stderr and the logging.logger to hosting process' freeorion_logger."""

    if not hasattr(redirect_logging_to_freeorion_logger, "only_redirect_once"):
        sys.stdout = _stdXLikeStream(logging.DEBUG)
        sys.stderr = _stdXLikeStream(logging.ERROR)
        print("Python stdout and stderr are redirected to ai process.")

        # thread and process information is already provided by boost logging framework
        # we can avoid some logging call overheads by turning this off
        logging.logThreads = 0
        logging.logProcesses = 0
        logging.logMultiprocessing = 0

        logger = logging.getLogger()
        logger.addHandler(_create_narrow_handler(logging.DEBUG))
        logger.addHandler(_create_narrow_handler(logging.INFO))
        logger.addHandler(_create_narrow_handler(logging.WARNING))
        logger.addHandler(_create_narrow_handler(logging.ERROR))
        logger.addHandler(_create_narrow_handler(logging.FATAL))
        logger.addHandler(_create_narrow_handler(logging.NOTSET))

        # Replace the system unhandled exception handler
        sys.excepthook = _unhandled_exception_hook

        logger.setLevel(initial_log_level)
        logger.info(
            "The python logger is initialized with a log level of %s" % logging.getLevelName(logger.getEffectiveLevel())
        )

        redirect_logging_to_freeorion_logger.only_redirect_once = True



from focs._effects import (
    AnyEmpire,
    Basic,
    EffectsGroup,
    HasTag,
    IsSource,
    MaxOf,
    OwnedBy,
    Planet,
    Population,
    SetVisibility,
    Source,
    ValueVisibility,
    VisibleToEmpire,
    WithinStarlaneJumps,
)


def PRECOGNITIVE_DETECTION(jumps: int):
    return [
        EffectsGroup(
            description="PRECOGNITIVE_DETECTION_DESC",
            scope=Planet()
            & ~VisibleToEmpire(empire=Source.Owner)
            & ~IsSource
            & ~OwnedBy(empire=Source.Owner)
            & WithinStarlaneJumps(jumps=jumps, condition=IsSource),
            activation=OwnedBy(affiliation=AnyEmpire),
            effects=SetVisibility(empire=Source.Owner, visibility=MaxOf("Visibility", Basic, ValueVisibility)),
        )
    ]


def TELEPATHIC_DETECTION(jumps: int) -> list:
    return [
        EffectsGroup(
            description="TELEPATHIC_DETECTION_DESC",
            scope=Planet()
            & ~Population(high=0)
            & ~OwnedBy(empire=Source.Owner)
            & ~VisibleToEmpire(empire=Source.Owner)
            & WithinStarlaneJumps(jumps=jumps, condition=IsSource)
            & ~IsSource,
            activation=OwnedBy(affiliation=AnyEmpire),
            effects=SetVisibility(empire=Source.Owner, visibility=MaxOf("Visibility", Basic, ValueVisibility)),
        )
    ]


def HAEMAESTHETIC_DETECTION(jumps: int) -> list:
    return [
        EffectsGroup(
            description="HAEMAESTHETIC_DETECTION_DESC",
            scope=Planet()
            & ~Population(high=0)
            & HasTag(name="ORGANIC")
            & ~OwnedBy(empire=Source.Owner)
            & ~VisibleToEmpire(empire=Source.Owner)
            & WithinStarlaneJumps(jumps=jumps, condition=IsSource)
            & ~IsSource,
            activation=OwnedBy(affiliation=AnyEmpire),
            effects=SetVisibility(empire=Source.Owner, visibility=MaxOf("Visibility", Basic, ValueVisibility)),
        )
    ]



# Policy Liberty doubles the stability malus from disliked buildings, policies and planetary foci
# Policy Conformance halves it
# Values are declared in common/named_values.focs.txt
from focs._effects import EmpireHasAdoptedPolicy, IsSource, NamedRealLookup, StatisticIf, Target

POLICY_DISLIKE_SCALING = (
    1.0
    * NamedRealLookup(name="PLC_LIBERTY_DISLIKE_FACTOR")
    ** (StatisticIf(float, condition=IsSource & EmpireHasAdoptedPolicy(empire=Target.Owner, name="PLC_LIBERTY")))
    * NamedRealLookup(name="PLC_CONFORMANCE_DISLIKE_FACTOR")
    ** (StatisticIf(float, condition=IsSource & EmpireHasAdoptedPolicy(empire=Target.Owner, name="PLC_CONFORMANCE")))
)



import freeorion as fo
import random
from collections import defaultdict

import universe_statistics
import universe_tables

# REPEAT_RATE along with calculate_number_of_specials_to_place determines if there are multiple
# specials in a single location.  There can only be at most 4 specials in a single location.
# The probabilites break down as follows:
# Count  Probability
# one    (1 - REPEAT_RATE[0])
# two    REPEAT_RATE[0] * (1 - REPEAT_RATE[1])
# three  REPEAT_RATE[0] * REPEAT_RATE[1] *(1 - REPEAT_RATE[2])
# four   REPEAT_RATE[0] * REPEAT_RATE[1] * REPEAT_RATE[2]
REPEAT_RATE = {1: 0.08, 2: 0.05, 3: 0.01, 4: 0.00}


def calculate_number_of_specials_to_place(objs):
    """Return a list of number of specials to be placed at each obj"""
    return [
        1
        if random.random() > REPEAT_RATE[1]
        else 2
        if random.random() > REPEAT_RATE[2]
        else 3
        if random.random() > REPEAT_RATE[3]
        else 4
        for _ in objs
    ]


def place_special(specials, obj):
    """
    Place at most a single special.
    Return the number of specials placed.
    """
    # Calculate the conditional probabilities that each special is
    # placed here given that a special will be placed here.
    probs = [fo.special_spawn_rate(sp) for sp in specials]
    total_prob = float(sum(probs))
    if total_prob == 0:
        # This shouldn't happen since special_spawn_rate > 0.0 is checked in distribute_specials()
        return 0
    thresholds = [x / total_prob for x in probs]

    chance = random.random()
    for threshold, special in zip(thresholds, specials):
        if chance > threshold:
            chance -= threshold
            continue

        fo.add_special(obj, special)
        print("Special", special, "added to", fo.get_name(obj))
        universe_statistics.specials_summary[special] += 1

        return 1
    return 0


# TODO Bug:  distribute_specials forward checks that a special can be
# placed, but it doesn't recursively check all previously placed
# specials against the new special.
def distribute_specials(specials_freq, universe_objects):  # noqa: C901
    """
    Adds start-of-game specials to universe objects.
    """
    # get basic chance for occurrence of specials from the universe tables
    base_chance = universe_tables.SPECIALS_FREQUENCY[specials_freq]
    if base_chance <= 0:
        return

    # get a list with all specials that have a spawn rate and limit both > 0 and a location condition defined
    # (no location condition means a special shouldn't get added at game start)
    specials = [
        sp
        for sp in fo.get_all_specials()
        if fo.special_spawn_rate(sp) > 0.0 and fo.special_spawn_limit(sp) > 0 and fo.special_has_location(sp)
    ]
    if not specials:
        return

    # dump a list of all specials meeting that conditions and their properties to the log
    print("Specials available for distribution at game start:")
    for special in specials:
        print(
            "... {:30}: spawn rate {:2.3f} / spawn limit {}".format(
                special, fo.special_spawn_rate(special), fo.special_spawn_limit(special)
            )
        )

    objects_needing_specials = [obj for obj in universe_objects if random.random() < base_chance]

    track_num_placed = {obj: 0 for obj in universe_objects}

    print(
        "Base chance for specials is {}. Placing specials on {} of {} ({:1.4f})objects".format(
            base_chance,
            len(objects_needing_specials),
            len(universe_objects),
            float(len(objects_needing_specials)) / len(universe_objects),
        )
    )

    obj_tuple_needing_specials = set(
        zip(
            objects_needing_specials,
            fo.objs_get_systems(objects_needing_specials),
            calculate_number_of_specials_to_place(objects_needing_specials),
        )
    )

    # Equal to the largest distance in WithinStarlaneJumps conditions
    # GALAXY_DECOUPLING_DISTANCE is used as follows.  For any two or more objects
    # at least GALAXY_DECOUPLING_DISTANCE appart you only need to check
    # fo.special_locations once and then you can place as many specials as possible,
    # subject to number restrictions.
    #
    # Organize the objects into sets where all objects are spaced GALAXY_DECOUPLING_DISTANCE
    # appart.  Place a special on each one.  Repeat until you run out of specials or objects.
    GALAXY_DECOUPLING_DISTANCE = 6

    while obj_tuple_needing_specials:
        systems_needing_specials = defaultdict(set)
        for obj, system, specials_count in obj_tuple_needing_specials:
            systems_needing_specials[system].add((obj, system, specials_count))

        print(f" Placing in {len(systems_needing_specials)} locations remaining.")

        # Find a list of candidates all spaced GALAXY_DECOUPLING_DISTANCE apart
        candidates = []
        while systems_needing_specials:
            random_sys = random.choice(list(systems_needing_specials.values()))
            member = random.choice(list(random_sys))
            obj, system, specials_count = member
            candidates.append(obj)
            obj_tuple_needing_specials.remove(member)
            if specials_count > 1:
                obj_tuple_needing_specials.add((obj, system, specials_count - 1))

            # remove all neighbors from the local pool
            for neighbor in fo.systems_within_jumps_unordered(GALAXY_DECOUPLING_DISTANCE, [system]):
                if neighbor in systems_needing_specials:
                    systems_needing_specials.pop(neighbor)

        print(
            "Caching specials_locations() at {} of {} remaining locations.".format(
                str(len(candidates)), str(len(obj_tuple_needing_specials) + len(candidates))
            )
        )
        # Get the locations at which each special can be placed
        locations_cache = {}
        for special in specials:
            # The fo.special_locations in the following line consumes most of the time in this
            # function.  Decreasing GALAXY_DECOUPLING_DISTANCE will speed up the whole
            # function by reducing the number of times this needs to be called.
            locations_cache[special] = set(fo.special_locations(special, candidates))

        # Attempt to apply a special to each candidate
        # by finding a special that can be applied to it and hasn't been added too many times
        for obj in candidates:
            # check if the spawn limit for this special has already been reached (that is, if this special
            # has already been added the maximal allowed number of times)
            specials = [s for s in specials if universe_statistics.specials_summary[s] < fo.special_spawn_limit(s)]
            if not specials:
                break

            # Find which specials can be placed at this one location
            local_specials = [sp for sp in specials if obj in locations_cache[sp]]
            if not local_specials:
                universe_statistics.specials_repeat_dist[0] += 1
                continue

            # All prerequisites and the test have been met, now add this special to this universe object.
            track_num_placed[obj] += place_special(local_specials, obj)

    for num_placed in track_num_placed.values():
        universe_statistics.specials_repeat_dist[num_placed] += 1



from focs._effects import (
    AddSpecial,
    Conditional,
    CreateShip,
    EffectsGroup,
    GalaxyMaxAIAggression,
    GalaxyMonsterFrequency,
    IsSource,
    Planet,
    Turn,
)
from focs._species import *
from macros.misc import MINIMUM_DISTANCE_EMPIRE_CHECK
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import SWAMP_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.planet_size import NOT_LARGE_PLANET
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import ULTIMATE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS

Species(
    name="SP_FIFTYSEVEN",
    description="SP_FIFTYSEVEN_DESC",
    gameplay_description="SP_FIFTYSEVEN_GAMEPLAY_DESC",
    native=True,
    tags=[
        "ORGANIC",
        "ARTISTIC",
        "NO_INDUSTRY",
        "ULTIMATE_RESEARCH",
        "NO_INFLUENCE",
        "BAD_POPULATION",
        "AVERAGE_SUPPLY",
        "GOOD_STEALTH",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        # /*[[HAS_INDUSTRY_FOCUS]]*/
        HAS_RESEARCH_FOCUS,
        # /*[[HAS_INFLUENCE_FOCUS]]*/
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "SHIMMER_SILK_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "SPICE_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "ELERIUM_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_DREAM_RECURSION",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_CENTRALIZATION",
        "PLC_ISOLATION",
    ],
    dislikes=[
        "BLD_CULTURE_LIBRARY",
        "BLD_SCANNING_FACILITY",
        "BLD_GENOME_BANK",
        "BLD_COLLECTIVE_NET",
        "BLD_ENCLAVE_VOID",
        "BLD_GATEWAY_VOID",
        "BLD_SHIPYARD_ENRG_COMP",
        "COMPUTRONIUM_SPECIAL",
        "PANOPTICON_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "CRYSTALS_SPECIAL",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_EXPLORATION",
        "PLC_ENGINEERING",
        "PLC_MARINE_RECRUITMENT",
        "PLC_TRAFFIC_CONTROL",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *ULTIMATE_RESEARCH,
        *NO_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_FIFTYSEVEN"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *GOOD_STEALTH,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *NOT_LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
        EffectsGroup(
            scope=IsSource,
            activation=Planet() & Turn(high=0) & (GalaxyMaxAIAggression >= 1) & (GalaxyMonsterFrequency >= 1),
            effects=[
                Conditional(
                    condition=MINIMUM_DISTANCE_EMPIRE_CHECK,
                    effects=[CreateShip(designname="SM_DRAGON")],
                    else_=[
                        AddSpecial(name="VOLCANIC_ASH_SLAVE_SPECIAL"),
                        AddSpecial(name="VOLCANIC_ASH_MASTER_SPECIAL"),
                    ],
                )
            ],
        ),
    ],
    environments=SWAMP_STANDARD_EP,
    graphic="icons/species/fifty-seven.png",
)



from focs._effects import GameRule

INDUSTRY_PER_POP = 0.2

STOCKPILE_PER_POP = 0.02

RESEARCH_PER_POP = 0.2

TROOPS_PER_POP = 0.2

TECH_COST_MULTIPLIER = GameRule(type=float, name="RULE_TECH_COST_FACTOR")

BUILDING_COST_MULTIPLIER = GameRule(type=float, name="RULE_BUILDING_COST_FACTOR")



import freeOrionAIInterface as fo
from functools import wraps

from aistate_interface import get_aistate


def cache_for_session(func):
    """
    Cache a function value for current session.

    Wraps only functions with hashable arguments.
    Use this only if the called function return value is constant throughout the game.
    """
    _cache = {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        key = (func, args, tuple(kwargs.items()))
        if key in _cache:
            return _cache[key]
        res = func(*args, **kwargs)
        _cache[key] = res
        return res

    wrapper._cache = _cache
    return wrapper


def cache_for_current_turn(func):
    """
    Cache a function value updated each turn.

    The cache is non-persistent through loading a game.
    Wraps only functions with hashable arguments.
    """
    _cache = {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        key = (func, args, tuple(kwargs.items()))
        this_turn = fo.currentTurn()
        if key in _cache and _cache[key][0] == this_turn:
            return _cache[key][1]
        res = func(*args, **kwargs)
        _cache[key] = (this_turn, res)
        return res

    wrapper._cache = _cache
    return wrapper


def cache_by_turn_persistent(func):
    """
    Cache a function value by turn, persistent through loading a game.

    It will also provides a history that may be analysed.
    The cache is keyed by the original function name. It only wraps functions without arguments.

    As the result is stored in AIstate, its type must be trusted by the savegame_codec module.
    """

    @wraps(func)
    def wrapper():
        if get_aistate() is None:
            return func()
        else:
            cache = get_aistate().misc.setdefault("caches", {}).setdefault(func.__name__, {})
            this_turn = fo.currentTurn()
            return cache[this_turn] if this_turn in cache else cache.setdefault(this_turn, func())

    return wrapper



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import INFERNO_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import GOOD_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS, GOOD_OFFENSE_TROOPS

Species(
    name="SP_UGMORS",
    description="SP_UGMORS_DESC",
    gameplay_description="SP_UGMORS_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "GOOD_INDUSTRY",
        "BAD_RESEARCH",
        "GOOD_INFLUENCE",
        "AVERAGE_SUPPLY",
        "GOOD_OFFENSE_TROOPS",
        "PEDIA_LITHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "FOCUS_STOCKPILE",
        "FOCUS_GROWTH",
        "MIMETIC_ALLOY_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "ELERIUM_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_CENTRALIZATION",
        "PLC_CONFORMANCE",
        "PLC_INDOCTRINATION",
        "PLC_TRAFFIC_CONTROL",
    ],
    dislikes=[
        "FOCUS_RESEARCH",
        "FOCUS_INFLUENCE",
        "BLD_BIOTERROR_PROJECTOR",
        "SUPERCONDUCTOR_SPECIAL",
        "SPICE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "PLC_LIBERTY",
        "PLC_MARINE_RECRUITMENT",
        "PLC_DIVINE_AUTHORITY",
        "PLC_CONFEDERATION",
    ],
    effectsgroups=[
        *GOOD_INDUSTRY,
        *BAD_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_UGMORS"),
        *AVERAGE_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        *GOOD_OFFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=INFERNO_NARROW_EP,
    graphic="icons/species/amorphous-06.png",
)
